#!/usr/bin/env python3

import os
import platform
import unittest
import pandas

from .. import available, get_url, load_example

from ..base import get_data_home

os_name = platform.system()


class Testexamples(unittest.TestCase):
    def test_available(self):
        examples = available()
        self.assertEqual(type(examples), pandas.core.frame.DataFrame)
        self.assertEqual(examples.shape, (98, 3))

    def test_data_home(self):
        pth = get_data_home()
        head, tail = os.path.split(pth)
        self.assertEqual(tail, "pysal")
        if os_name == "Linux":
            heads = head.split("/")
            self.assertEqual(heads[1], "home")
            self.assertEqual(heads[-1], "share")
            self.assertEqual(heads[-2], ".local")
        elif os_name == "Darwin":
            heads = head.split("/")
            self.assertEqual(heads[1], "Users")
            self.assertEqual(heads[-1], "Application Support")
            self.assertEqual(heads[-2], "Library")
        elif os_name == "Windows":
            heads = head.split("\\")
            self.assertEqual(heads[1], "Users")
            self.assertEqual(heads[-2], "Local")
            self.assertEqual(heads[-3], "AppData")

    def test_get_url(self):
        self.assertEqual(get_url("10740"), None)
        url = "https://geodacenter.github.io/data-and-lab//data/baltimore.zip"
        self.assertEqual(get_url("Baltimore"), url)

    def test_load_example(self):
        taz = load_example("taz")
        flist = taz.get_file_list()
        self.assertEqual(len(flist), 4)


suite = unittest.TestLoader().loadTestsFromTestCase(Testexamples)

if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite)



# pylint: disable-msg=C0103, C0301
"""
Pure Python implementation of RTree spatial index.

Adaptation of
http://code.google.com/p/pyrtree/

R-tree.
see doc/ref/r-tree-clustering-split-algo.pdf
"""

__author__ = "Sergio J. Rey"

__all__ = ["RTree", "Rect", "Rtree"]


import array
import numpy
import random
import time

MAXCHILDREN = 10
MAX_KMEANS = 5
BUFFER = numpy.finfo(float).eps


class Rect(object):
    """A rectangle class that stores an axis aligned rectangle and two flags
    (swapped_x and swapped_y). The flags are stored implicitly via swaps in
    the order of minx/y and maxx/y.
    """

    __slots__ = ("x", "y", "xx", "yy", "swapped_x", "swapped_y")

    def __getstate__(self) -> tuple:
        return (self.x, self.y, self.xx, self.yy, self.swapped_x, self.swapped_y)

    def __setstate__(self, state: tuple):
        self.x, self.y, self.xx, self.yy, self.swapped_x, self.swapped_y = state

    def __init__(self, minx: float, miny: float, maxx: float, maxy: float):

        self.swapped_x = maxx < minx
        self.swapped_y = maxy < miny
        self.x = minx
        self.y = miny
        self.xx = maxx
        self.yy = maxy

        if self.swapped_x:
            self.x, self.xx = maxx, minx
        if self.swapped_y:
            self.y, self.yy = maxy, miny

    def coords(self) -> tuple:
        """Return the coordinates of the rectangle."""

        return self.x, self.y, self.xx, self.yy

    def overlap(self, orect):
        """Return the overlapping area of two rectangles.
        
        Parameters
        ----------
        orect : libpysal.cg.Rect
            Another rectangle.
        
        Returns
        -------
        overlapping_area : float
            The area of the overlap between ``orect`` and ``self``.
        
        """

        overlapping_area = self.intersect(orect).area()

        return overlapping_area

    def write_raw_coords(self, toarray, idx: int):
        """Write the raw coordinates of the rectangle."""

        toarray[idx] = self.x
        toarray[idx + 1] = self.y
        toarray[idx + 2] = self.xx
        toarray[idx + 3] = self.yy
        if self.swapped_x:
            toarray[idx] = self.xx
            toarray[idx + 2] = self.x
        if self.swapped_y:
            toarray[idx + 1] = self.yy
            toarray[idx + 3] = self.y

    def area(self) -> float:
        """Calculate the area of the rectangle."""

        w = self.xx - self.x
        h = self.yy - self.y

        return w * h

    def extent(self) -> tuple:
        """Return the extent of the rectangle in the form: (minx, minx, width, height)."""

        x = self.x
        y = self.y

        return (x, y, self.xx - x, self.yy - y)

    def grow(self, amt=None, sf=0.5):
        """Grow the bounds of a rectangle.
        
        Parameters
        ----------
        amt : float
            The amount to grow the rectangle. Default is ``None``, which
            triggers the value of ``BUFFER``.
        sf : float
            The scale factor for ``amt``. Default is ``0.5``.
            
        Returns
        -------
        rect : libpysal.cg.Rect
            A new rectangle grown by ``amt`` and scaled by ``sf``.
        
        """

        if not amt:
            amt = BUFFER
        a = amt * sf
        rect = Rect(self.x - a, self.y - a, self.xx + a, self.yy + a)

        return rect

    def intersect(self, o):
        """Find the intersection of two rectangles.
        
        Parameters
        ----------
        o : libpysal.cg.Rect
            Another rectangle.
            
        Returns
        -------
        intersection : {libpysal.cg.NullRect, libpysal.cg.Rect}
            The intersecting part of ``o`` and ``self``.

        """

        intersection = None

        if self is NullRect:
            intersection = NullRect
        elif o is NullRect:
            intersection = NullRect

        if not intersection:

            nx, ny = max(self.x, o.x), max(self.y, o.y)
            nx2, ny2 = min(self.xx, o.xx), min(self.yy, o.yy)
            w, h = nx2 - nx, ny2 - ny

            if w <= 0 or h <= 0:
                intersection = NullRect
            else:
                intersection = Rect(nx, ny, nx2, ny2)

        return intersection

    def does_contain(self, o):
        """Check whether the rectangle contains the other rectangle.
        
        Parameters
        ----------
        o : libpysal.cg.Rect
            Another rectangle.
        
        Returns
        -------
        dc : bool
            ``True`` if ``self`` contains ``o`` otherwise ``False``.
        
        """

        dc = self.does_containpoint((o.x, o.y)) and self.does_containpoint((o.xx, o.yy))

        return dc

    def does_intersect(self, o):
        """Check whether the rectangles interect.
        
        Parameters
        ----------
        o : libpysal.cg.Rect
            Another rectangle.
        
        Returns
        -------
        dcp : bool
            ``True`` if ``self`` intersects ``o`` otherwise ``False``.
        
        """

        di = self.intersect(o).area() > 0

        return di

    def does_containpoint(self, p):
        """Check whether the rectangle contains a point or not.
        
        Parameters
        ----------
        p : libpysal.cg.Point
            A point.
        
        Returns
        -------
        dcp : bool
            ``True`` if ``self`` contains ``p`` otherwise ``False``.
        
        """

        x, y = p

        dcp = x >= self.x and x <= self.xx and y >= self.y and y <= self.yy

        return dcp

    def union(self, o):
        """Union two rectangles.
        
        Parameters
        ----------
        o : libpysal.cg.Rect
            Another rectangle.
        
        Returns
        -------
        res : libpysal.cg.Rect
            The union of ``o`` and ``self``.

        """

        if o is NullRect:
            res = Rect(self.x, self.y, self.xx, self.yy)
        elif self is NullRect:
            res = Rect(o.x, o.y, o.xx, o.yy)
        else:
            x = self.x
            y = self.y
            xx = self.xx
            yy = self.yy
            ox = o.x
            oy = o.y
            oxx = o.xx
            oyy = o.yy

            nx = x if x < ox else ox
            ny = y if y < oy else oy
            nx2 = xx if xx > oxx else oxx
            ny2 = yy if yy > oyy else oyy

            res = Rect(nx, ny, nx2, ny2)

        return res

    def union_point(self, o):
        """Union the rectangle and a point
        
        Parameters
        ----------
        o : libpysal.cg.Point
            A point.
        
        Returns
        -------
        res : libpysal.cg.Rect
            The union of ``o`` and ``self``.

        """

        x, y = o
        res = self.union(Rect(x, y, x, y))

        return res

    def diagonal_sq(self) -> float:
        """Calculate the squared diagonal of the rectangle."""

        if self is NullRect:
            diag_sq = 0.0

        else:
            w = self.xx - self.x
            h = self.yy - self.y
            diag_sq = w * w + h * h

        return diag_sq

    def diagonal(self) -> float:
        """Calculate the diagonal of the rectangle."""

        return numpy.sqrt(self.diagonal_sq())


NullRect = Rect(0.0, 0.0, 0.0, 0.0)
NullRect.swapped_x = False
NullRect.swapped_y = False


def union_all(kids):
    """Create union of all child rectangles.
    
    Parameters
    ----------
    kids : list
        A list of ``libpysal.cg._NodeCursor`` objects.
    
    Returns
    -------
    cur : {libpysal.cg.Rect, libpysal.cg.NullRect}
        The unioned result of all child rectangles.
    
    """

    cur = NullRect
    for k in kids:
        cur = cur.union(k.rect)

    assert False == cur.swapped_x

    return cur


def Rtree():
    return RTree()


class RTree(object):
    """An RTree for efficiently querying space based on intersecting rectangles.
    
    Attributes
    ----------
    count : int
        The number of nodes in the tree.
    stats : dict
        Tree generation statistics.
    leaf_count : int
        The number of leaves (objects) in the tree.
    rect_pool : array.array
        The pool of rectangles in the tree in the form
        :math:`[ax1, ay1, ax2, ay2, bx1, by1, bx2, by2, ..., nx1, ny1, nx2, ny2]`
        where the first set of 4 coordinates is the bounding box of the root node
        and each successive set of 4 coordinates is the bounding box of a leaf node.
    node_pool : array.array
        The pool of node IDs in the tree.
    leaf_pool : list
        The pool of leaf objects in the tree.
    cursor : libpysal.cg._NodeCursor
        The non-root node and all its children.
    
    Examples
    --------
    
    Instantiate an ``RTree``.
    
    >>> from libpysal.cg import RTree, Chain
    >>> segments = [
    ...     [(0.0, 1.5), (1.5, 1.5)],
    ...     [(1.5, 1.5), (3.0, 1.5)],
    ...     [(1.5, 1.5), (1.5, 0.0)],
    ...     [(1.5, 1.5), (1.5, 3.0)]
    ... ]
    >>> segments = [Chain([p1, p2]) for p1, p2 in segments]
    >>> rt = RTree()
    >>> for segment in segments:
    ...     rt.insert(segment, Rect(*segment.bounding_box).grow(sf=10.))
    
    Examine the tree generation statistics. The statistics here
    are all 0 due to the simple structure of the tree in this example.
    
    >>> rt.stats
    {'overflow_f': 0,
     'avg_overflow_t_f': 0.0,
     'longest_overflow': 0.0,
     'longest_kmeans': 0.0,
     'sum_kmeans_iter_f': 0,
     'count_kmeans_iter_f': 0,
     'avg_kmeans_iter_f': 0.0}
    
    Examine the number of nodes and leaves.
    There five nodes and four leaves (the root plus its four children).
    
    >>> rt.count, rt.leaf_count
    (5, 4)
    
    The pool of nodes are the node IDs in the tree.
    
    >>> rt.node_pool
    array('L', [0, 4, 0, 0, 1, 1, 2, 2, 3, 3])
    
    The pool of leaves are the geometric objects that were inserted into the tree.
    
    >>> rt.leaf_pool[0].vertices
    [(0.0, 1.5), (1.5, 1.5)]
    
    The pool of rectangles are the bounds of partitioned space in the tree.
    Examine the first one.
    
    >>> rt.rect_pool[:4]
    array('d', [-2.220446049250313e-15, -2.220446049250313e-15, 3.000000000000002, 3.000000000000002])
    
    Add the bounding box of a leaf to the tree manually.
    
    >>> rt.add(Chain(((2,2), (4,4))), (2,2,4,4))
    >>> rt.count, rt.leaf_count
    (6, 5)
    
    Query the tree for an intersection. One object is contained in this query.
    
    >>> rt.intersection([.4, 2.1, .9, 2.6])[0].vertices
    [(0.5, 2), (1, 2.5)]
    
    Query the tree with a much larger box. All objects are contained in this query.
    
    >>> len(rt.intersection([-1, -1, 4, 4])) == rt.leaf_count
    True
    
    Query the tree with box outside the tree objects.
    No objects are contained in this query.
    
    >>> rt.intersection([5, 5, 6, 6])
    []
    
    """

    def __init__(self):

        self.count = 0
        self.stats = {
            "overflow_f": 0,
            "avg_overflow_t_f": 0.0,
            "longest_overflow": 0.0,
            "longest_kmeans": 0.0,
            "sum_kmeans_iter_f": 0,
            "count_kmeans_iter_f": 0,
            "avg_kmeans_iter_f": 0.0,
        }

        # This round: not using objects directly -- they take up too much memory, and
        #   efficiency goes down the toilet (obviously) if things start to page. Less
        #   obviously: using object graph directly leads to really long GC pause times,
        #   too. Instead, it uses pools of arrays:
        self.count = 0
        self.leaf_count = 0
        self.rect_pool = array.array("d")
        self.node_pool = array.array("L")

        # leaf objects.
        self.leaf_pool = []

        self.cursor = _NodeCursor.create(self, NullRect)

    def _ensure_pool(self, idx: int):
        """Ensure sufficient slots in rectangle and node pools."""

        bb_len, pool_slot = 4, [0]
        node_len = int(bb_len / 2)
        if len(self.rect_pool) < (bb_len * idx):
            self.rect_pool.extend(pool_slot * bb_len)
            self.node_pool.extend(pool_slot * node_len)

    def insert(self, o, orect):
        """Insert an object and its bounding box into the tree.
        
        Parameters
        ----------
        o : libpysal.cg.{Point, Chain, Rectangle, Polygon}
            The object to insert into the tree.
        orect : ibpysal.cg.Rect
            The object's bounding box.
        
        """

        self.cursor.insert(o, orect)
        assert self.cursor.index == 0

    def query_rect(self, r):
        """Query a rectangle.
        
        Parameters
        ----------
        r : {tuple, libpysal.cg.Point}
            The bounding box of the rectangle in question;
            a :math:`(minx,miny,maxx,maxy)` set of coordinates.
        
        Yields
        ------
        x : generator
            ``libpysal.cg._NodeCursor`` objects.
        """

        for x in self.cursor.query_rect(r):
            yield x

    def query_point(self, p):
        """Query a point.
        
        Parameters
        ----------
        p : {tuple, libpysal.cg.Point}
            The point in question; an :math:`(x,y)` coordinate.
        
        Yields
        ------
        x : generator
            ``libpysal.cg._NodeCursor`` objects.
        """

        for x in self.cursor.query_point(p):
            yield x

    def walk(self, pred):
        """Walk the tree structure with ``pred`` (a function)."""

        return self.cursor.walk(pred)

    def intersection(self, boundingbox):
        """Query for an intersection between leaves in the ``RTree``
        and the bounding box of an object.
        
        Parameters
        ----------
        boundingbox : list
            The bounding box: ``[minx, miny, maxx, maxy]``.
        
        Returns
        -------
        objs : list
            A list of objects whose bounding
            boxes intersect with the query bounding box.

        """

        # grow the bounding box slightly to handle coincident edges
        qr = Rect(*boundingbox).grow(sf=10.0)

        objs = [r.leaf_obj() for r in self.query_rect(qr) if r.is_leaf()]

        return objs

    def add(self, id, boundingbox):
        """Add the bounding box of a leaf to the ``RTree`` manually with a specified ID.

        Parameters
        ----------
        id : int
            An object id.
        boundingbox : list
            The bounding box: ``[minx, miny, maxx, maxy]``.
        
        """

        self.cursor.insert(id, Rect(*boundingbox))


class _NodeCursor(object):
    """An internal class for keeping track of, and reorganizing,
    the structure and composition of the ``RTree``.
    
    Parameters
    ----------
    rooto : libpysal.cg.{Point, Chain, Rectangle, Polygon}
        The object from which the node will be generated.
    index : int
        The ID of the node.
    rect : libpysal.cg.Rect
        The bounding rectangle of the leaf object.
    first_child : int
        The ID of the first child of the node.
    next_sibling : int
        The ID of the sibling of the node.
    
    Attributes
    ----------
    root : libpysal.cg.RTree
        The root node of the tree.
    npool : array.array
        See ``RTree.node_pool``.
    rpool : array.array
        See ``RTree.rect_pool``.
    
    """

    @classmethod
    def create(cls, rooto, rect):
        """Create a node in the tree structure.
        
        Parameters
        ----------
        rooto : libpysal.cg.{Point, Chain, Rectangle, Polygon}
            The object from which the node will be generated.
        index : int
            The ID of the node.
        rect : libpysal.cg.Rect
            The bounding rectangle of the leaf object.
        
        Returns
        -------
        retv : libpysal.cg._NodeCursor
            The generated node.
        
        """

        idx = rooto.count
        rooto.count += 1

        rooto._ensure_pool(idx + 1)

        retv = _NodeCursor(rooto, idx, rect, 0, 0)

        retv._save_back()

        return retv

    @classmethod
    def create_with_children(cls, children, rooto):
        """Create a non-leaf node in the tree structure.
        
        Parameters
        ----------
        children : list
            The child nodes of the node to be generated
        rooto : libpysal.cg.{Point, Chain, Rectangle, Polygon}
            The object from which the node will be generated.
        
        Returns
        -------
        nc : libpysal.cg._NodeCursor
            The generated node with children.
        
        """
        rect = union_all([c for c in children])
        nr = Rect(rect.x, rect.y, rect.xx, rect.yy)

        assert not rect.swapped_x
        nc = _NodeCursor.create(rooto, rect)
        nc._set_children(children)

        assert not nc.is_leaf()
        return nc

    @classmethod
    def create_leaf(cls, rooto, leaf_obj, leaf_rect):
        """Create a leaf node in the tree structure.
        
        Parameters
        ----------
        rooto : libpysal.cg.{Point, Chain, Rectangle, Polygon}
            The object from which the node will be generated.
        leaf_obj : libpysal.cg.{Point, Chain, Rectangle, Polygon}
            The leaf object.
        leaf_rect : libpysal.cg.Rect
            The bounding rectangle of the leaf object.
        
        Returns
        -------
        res : libpysal.cg._NodeCursor
            The generated leaf node.
        
        """

        rect = Rect(leaf_rect.x, leaf_rect.y, leaf_rect.xx, leaf_rect.yy)

        # Mark as leaf by setting the xswap flag.
        rect.swapped_x = True
        res = _NodeCursor.create(rooto, rect)
        idx = res.index
        res.first_child = rooto.leaf_count
        rooto.leaf_count += 1
        res.next_sibling = 0
        rooto.leaf_pool.append(leaf_obj)
        res._save_back()
        res._become(idx)

        assert res.is_leaf()
        return res

    __slots__ = (
        "root",
        "npool",
        "rpool",
        "index",
        "rect",
        "next_sibling",
        "first_child",
    )

    def __getstate__(self) -> tuple:
        return (
            self.root,
            self.npool,
            self.rpool,
            self.index,
            self.rect,
            self.next_sibling,
            self.first_child,
        )

    def __setstate__(self, state: tuple):
        (
            self.root,
            self.npool,
            self.rpool,
            self.index,
            self.rect,
            self.next_sibling,
            self.first_child,
        ) = state

    def __init__(self, rooto, index, rect, first_child, next_sibling):

        self.root = rooto
        self.rpool = rooto.rect_pool
        self.npool = rooto.node_pool
        self.index = index
        self.rect = rect
        self.next_sibling = next_sibling
        self.first_child = first_child

    def walk(self, predicate):
        """Walk the tree structure with ``predicate`` (a function)."""

        if predicate(self, self.leaf_obj()):
            yield self
            if not self.is_leaf():
                for c in self.children():
                    for cr in c.walk(predicate):
                        yield cr

    def query_rect(self, r):
        """Yield objects that intersect with the rectangle (``r``)."""

        def p(o, x):
            return r.does_intersect(o.rect)

        for rr in self.walk(p):
            yield rr

    def query_point(self, point):
        """Yield objects that intersect with the point (``point``)."""

        def p(o, x):
            return o.rect.does_containpoint(point)

        for rr in self.walk(p):
            yield rr

    def lift(self):
        """Promote a node to (potentially) rearrange the
        tree structure for optimal clustering.
        
        Called from ``_NodeCursor._balance()``.
        
        Returns
        -------
        lifted : libpysal.cg._NodeCursor
            The lifted node.
        
        """

        lifted = _NodeCursor(
            self.root, self.index, self.rect, self.first_child, self.next_sibling
        )

        return lifted

    def _become(self, index: int):
        """Have ``self`` become node ``index``."""

        recti = index * 4
        nodei = index * 2
        rp = self.rpool
        x = rp[recti]
        y = rp[recti + 1]
        xx = rp[recti + 2]
        yy = rp[recti + 3]

        if x == 0.0 and y == 0.0 and xx == 0.0 and yy == 0.0:
            self.rect = NullRect
        else:
            self.rect = Rect(x, y, xx, yy)

        self.next_sibling = self.npool[nodei]
        self.first_child = self.npool[nodei + 1]
        self.index = index

    def is_leaf(self) -> bool:
        """Return ``True`` if the node is a leaf, otherwise ``False``."""

        return self.rect.swapped_x

    def has_children(self) -> bool:
        """Return ``True`` if the node has children, otherwise ``False``."""

        return not self.is_leaf() and 0 != self.first_child

    def holds_leaves(self) -> bool:
        """Return ``True`` if the node holds leaves, otherwise ``False``."""

        if 0 == self.first_child:
            return True
        else:
            return self.has_children() and self.get_first_child().is_leaf()

    def get_first_child(self):
        """Get the first child of a node.
        
        Returns
        -------
        c : libpysal.cg._NodeCursor
            The first child of the specified node.
        
        """

        fc = self.first_child
        c = _NodeCursor(self.root, 0, NullRect, 0, 0)
        c._become(self.first_child)

        return c

    def leaf_obj(self):
        """Return the leaf object if the node is a leaf, other return ``None``."""

        if self.is_leaf():
            return self.root.leaf_pool[self.first_child]
        else:
            return None

    def _save_back(self):
        """Save a node back into the tree structure."""

        rp = self.rpool
        recti = self.index * 4
        nodei = self.index * 2

        if self.rect is not NullRect:
            self.rect.write_raw_coords(rp, recti)
        else:
            rp[recti] = 0
            rp[recti + 1] = 0
            rp[recti + 2] = 0
            rp[recti + 3] = 0

        self.npool[nodei] = self.next_sibling
        self.npool[nodei + 1] = self.first_child

    def nchildren(self) -> int:
        """The number of children nodes."""

        i = self.index
        c = 0
        for x in self.children():
            c += 1

        return c

    def insert(self, leafo, leafrect):
        """Insert a leaf object into the tree. See
        ``RTree.insert(o, orect)`` for parameter description.
        
        """

        index = self.index

        # tail recursion, made into loop:
        while True:
            if self.holds_leaves():
                self.rect = self.rect.union(leafrect)
                self._insert_child(_NodeCursor.create_leaf(self.root, leafo, leafrect))

                self._balance()

                # done: become the original again
                self._become(index)
                return
            else:
                # Not holding leaves, move down a level in the tree:

                # ----------------------
                # Micro-optimization:
                #   inlining union() calls -- logic is:
                #       ignored, child = min(
                #           [
                #               ((c.rect.union(leafrect)).area() - c.rect.area(),c.index)
                #               for c in self.children()
                #           ]
                #       )
                child = None
                minarea = -1.0
                for c in self.children():
                    x, y, xx, yy = c.rect.coords()
                    lx, ly, lxx, lyy = leafrect.coords()
                    nx = x if x < lx else lx
                    nxx = xx if xx > lxx else lxx
                    ny = y if y < ly else ly
                    nyy = yy if yy > lyy else lyy
                    a = (nxx - nx) * (nyy - ny)
                    if minarea < 0 or a < minarea:
                        minarea = a
                        child = c.index
                # End micro-optimization
                # ----------------------

                self.rect = self.rect.union(leafrect)
                self._save_back()
                # recurse.
                self._become(child)

    def _balance(self):
        """Balance the leaf layout where possible through ``k_means_cluster()``
        and ``silhouette_coeff()`` for (heuristically) optimal clusterings of
        nodes in the tree structure after the child count of a node has grown
        past the maximum allowed number (see ``MAXCHILDREN``).
        
        Called from ``_NodeCursor.insert()``.
        """

        if self.nchildren() <= MAXCHILDREN:
            return

        t = time.process_time()

        cur_score = -10

        s_children = [c.lift() for c in self.children()]

        clusterings = [
            k_means_cluster(self.root, k, s_children) for k in range(2, MAX_KMEANS)
        ]
        score, bestcluster = max([(silhouette_coeff(c), c) for c in clusterings])

        # generate the (heuristically) optimally-balanced cluster of nodes
        nodes = [
            _NodeCursor.create_with_children(c, self.root)
            for c in bestcluster
            if len(c) > 0
        ]

        self._set_children(nodes)

        dur = time.process_time() - t
        c = float(self.root.stats["overflow_f"])
        oa = self.root.stats["avg_overflow_t_f"]
        self.root.stats["avg_overflow_t_f"] = (dur / (c + 1.0)) + (c * oa / (c + 1.0))
        self.root.stats["overflow_f"] += 1
        self.root.stats["longest_overflow"] = max(
            self.root.stats["longest_overflow"], dur
        )

    def _set_children(self, cs: list):
        """Set up the (new/altered) leaf tree structure.
        
        Called from ``_NodeCursor.create_with_children()``
        and ``_NodeCursor._balance()``.
        
        """

        self.first_child = 0

        if 0 == len(cs):
            return

        pred = None
        for c in cs:
            if pred is not None:
                pred.next_sibling = c.index
                pred._save_back()
            if 0 == self.first_child:
                self.first_child = c.index
            pred = c
        pred.next_sibling = 0
        pred._save_back()
        self._save_back()

    def _insert_child(self, c):
        """Internal function for child node insertion. 
        Called from ``_NodeCursor.insert()``.
        
        Parameters
        ----------
        c : libpysal.cg._NodeCursor
            A child ``libpysal.cg._NodeCursor`` object.
        
        """

        c.next_sibling = self.first_child
        self.first_child = c.index
        c._save_back()
        self._save_back()

    def children(self):
        """Yield the children of a node."""

        if 0 == self.first_child:
            return

        idx = self.index
        fc = self.first_child
        ns = self.next_sibling
        r = self.rect

        self._become(self.first_child)
        while True:
            yield self
            if 0 == self.next_sibling:
                break
            else:
                self._become(self.next_sibling)

        # Go back to becoming the same node we were.
        # self._become(idx)
        self.index = idx
        self.first_child = fc
        self.next_sibling = ns
        self.rect = r


def avg_diagonals(node, onodes):
    """Calculate the mean diagonals.
    
    Parameters
    ----------
    node : libpysal.cg._NodeCursor
        The target node in question.
    onodes : ist
        A list of ``libpysal.cg._NodeCursor`` objects.
    
    Returns
    -------
    diag_avg : float
        The mean diagonal distance of ``node`` and ``onodes``.
    
    """

    nidx = node.index
    sv = 0.0
    diag = 0.0
    memo_tab = {}

    for onode in onodes:
        k1 = (nidx, onode.index)
        k2 = (onode.index, nidx)

        if k1 in memo_tab:
            diag = memo_tab[k1]
        elif k2 in memo_tab:
            diag = memo_tab[k2]
        else:
            diag = node.rect.union(onode.rect).diagonal()
            memo_tab[k1] = diag

        sv += diag

    diag_avg = sv / len(onodes)

    return diag_avg


def silhouette_w(node, cluster, next_closest_cluster):
    """Calculate a silhouette score between a certain node and 2 clusters:
    
    Parameters
    ----------
    node : libpysal.cg._NodeCursor
        The target node in question.
    cluster : list
        A list of ``libpysal.cg._NodeCursor`` objects.
    next_closest_cluster : list
        Another list of ``libpysal.cg._NodeCursor`` objects.
    
    Returns
    -------
    silw : float
        The silhouette score between ``{node, cluster}``
        and ``{node, next_closest_cluster}``.
    
    """

    ndist = avg_diagonals(node, cluster)
    sdist = avg_diagonals(node, next_closest_cluster)

    silw = (sdist - ndist) / max(sdist, ndist)

    return silw


def silhouette_coeff(clustering):
    """Calculate how well defined the clusters are. A score of ``1`` indicates
    the clusters are well defined, a score of ``0`` indicates the clusters are
    undefined, and a score of ``-1`` indicates the clusters are defined
    incorrectly.
    
    Parameters
    ----------
    clustering : list
        A list of ``libpysal.cg._NodeCursor`` objects.
    
    Returns
    -------
    silcoeff : float
        Score for how well defined the clusters are.
        
    """

    # special case for a clustering of 1.0
    if len(clustering) == 1:
        silcoeff = 1.0

    else:
        coeffs = []
        for cluster in clustering:
            others = [c for c in clustering if c is not cluster]
            others_cntr = [center_of_gravity(c) for c in others]
            ws = [
                silhouette_w(node, cluster, others[closest(others_cntr, node)])
                for node in cluster
            ]
            cluster_coeff = sum(ws) / len(ws)
            coeffs.append(cluster_coeff)

        silcoeff = sum(coeffs) / len(coeffs)

    return silcoeff


def center_of_gravity(nodes):
    """Find the center of gravity of multiple nodes.
    
    Parameters
    ----------
    nodes : list
        A list of ``libpysal.cg.RTree`` and ``libpysal.cg._NodeCursor`` objects.
    
    Returns
    -------
    cog : float
        The center of gravity of multiple nodes.
    
    """

    totarea = 0.0
    xs, ys = 0, 0
    for n in nodes:
        if n.rect is not NullRect:
            x, y, w, h = n.rect.extent()
            a = w * h
            xs = xs + (a * (x + (0.5 * w)))
            ys = ys + (a * (y + (0.5 * h)))
            totarea = totarea + a

    cog = (xs / totarea), (ys / totarea)

    return cog


def closest(centroids, node):
    """Find the closest controid to the node's center of gravity.
    
    Parameters
    ----------
    centroids : list
        A list of (x, y) coordinates for the center of other clusters.
    node : libpysal.cg_NodeCursor
        A ``libpysal.cg._NodeCursor`` instance.
    
    Returns
    -------
    ridx : int
        The index of the nearest centroid of other cluster.

    """

    x, y = center_of_gravity([node])
    dist = -1
    ridx = -1

    for (i, (xx, yy)) in enumerate(centroids):
        dsq = ((xx - x) ** 2) + ((yy - y) ** 2)
        if -1 == dist or dsq < dist:
            dist = dsq
            ridx = i

    return ridx


def k_means_cluster(root, k, nodes):
    """Find ``k`` clusters.
    
    Parameters
    ----------
    root : libpysal.cg.RTree
        An ``libpysal.cg.RTree`` instance.
    k : int
        The number clusters to find.
    nodes : list
        A list of ``libpysal.cg.RTree`` and ``libpysal.cg._NodeCursor`` objects.
    
    Returns
    -------
    clusters : list
        Updated versions of ``nodes`` defining new clusters.
    
    """

    t = time.process_time()
    if len(nodes) <= k:
        clusters = [[n] for n in nodes]
        return clusters

    ns = list(nodes)
    root.stats["count_kmeans_iter_f"] += 1

    # Initialize: take n random nodes.
    # random.shuffle(ns)

    cluster_starts = ns[:k]
    cluster_centers = [center_of_gravity([n]) for n in ns[:k]]

    # Loop until stable:
    while True:
        root.stats["sum_kmeans_iter_f"] += 1
        clusters = [[] for c in cluster_centers]

        for n in ns:
            idx = closest(cluster_centers, n)
            clusters[idx].append(n)

        # FIXME HACK TODO: is it okay for there to be empty clusters?
        clusters = [c for c in clusters if len(c) > 0]

        for c in clusters:
            if len(c) == 0:
                print("Error....")
                print(("Nodes: %d, centers: %s." % (len(ns), repr(cluster_centers))))

            assert len(c) > 0

        rest = ns
        first = False

        new_cluster_centers = [center_of_gravity(c) for c in clusters]
        if new_cluster_centers == cluster_centers:
            root.stats["avg_kmeans_iter_f"] = float(
                root.stats["sum_kmeans_iter_f"] / root.stats["count_kmeans_iter_f"]
            )
            root.stats["longest_kmeans"] = max(
                root.stats["longest_kmeans"], (time.process_time() - t)
            )
            return clusters
        else:
            cluster_centers = new_cluster_centers



"""Unit test for util.py"""
from ..util import lat2W, nonplanar_neighbors, fuzzy_contiguity
from .. import util
from ..weights import W, WSP
from ..distance import DistanceBand, KNN
from ..contiguity import Queen, Rook
from ...io.fileio import FileIO as psopen
from ... import examples
import numpy as np
import scipy
import unittest


try:
    import geopandas as gpd

    HAS_GEOPANDAS = True
except:
    HAS_GEOPANDAS = False


class Testutil(unittest.TestCase):
    def setUp(self):
        self.w = Rook.from_shapefile(examples.get_path("10740.shp"))

        self.rio = examples.load_example("Rio Grande do Sul")

    def test_lat2W(self):
        w9 = lat2W(3, 3)
        self.assertEqual(w9.pct_nonzero, 29.62962962962963)
        self.assertEqual(w9[0], {1: 1.0, 3: 1.0})
        self.assertEqual(w9[3], {0: 1.0, 4: 1.0, 6: 1.0})

    def test_lat2SW(self):
        w9 = util.lat2SW(3, 3)
        rows, cols = w9.shape
        n = rows * cols
        self.assertEqual(w9.nnz, 24)
        pct_nonzero = w9.nnz / float(n)
        self.assertEqual(pct_nonzero, 0.29629629629629628)
        data = w9.todense().tolist()
        self.assertEqual(data[0], [0, 1, 0, 1, 0, 0, 0, 0, 0])
        self.assertEqual(data[1], [1, 0, 1, 0, 1, 0, 0, 0, 0])
        self.assertEqual(data[2], [0, 1, 0, 0, 0, 1, 0, 0, 0])
        self.assertEqual(data[3], [1, 0, 0, 0, 1, 0, 1, 0, 0])
        self.assertEqual(data[4], [0, 1, 0, 1, 0, 1, 0, 1, 0])
        self.assertEqual(data[5], [0, 0, 1, 0, 1, 0, 0, 0, 1])
        self.assertEqual(data[6], [0, 0, 0, 1, 0, 0, 0, 1, 0])
        self.assertEqual(data[7], [0, 0, 0, 0, 1, 0, 1, 0, 1])
        self.assertEqual(data[8], [0, 0, 0, 0, 0, 1, 0, 1, 0])

    def test_block_weights(self):
        regimes = np.ones(25)
        regimes[list(range(10, 20))] = 2
        regimes[list(range(21, 25))] = 3
        regimes = np.array(
            [
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                2.0,
                2.0,
                2.0,
                2.0,
                2.0,
                2.0,
                2.0,
                2.0,
                2.0,
                2.0,
                1.0,
                3.0,
                3.0,
                3.0,
                3.0,
            ]
        )
        w = util.block_weights(regimes)
        ww0 = [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        self.assertEqual(w.weights[0], ww0)
        wn0 = [1, 2, 3, 4, 5, 6, 7, 8, 9, 20]
        self.assertEqual(w.neighbors[0], wn0)
        regimes = ["n", "n", "s", "s", "e", "e", "w", "w", "e"]
        w = util.block_weights(regimes)
        wn = {
            0: [1],
            1: [0],
            2: [3],
            3: [2],
            4: [5, 8],
            5: [4, 8],
            6: [7],
            7: [6],
            8: [4, 5],
        }
        self.assertEqual(w.neighbors, wn)
        ids = ["id-%i" % i for i in range(len(regimes))]
        w = util.block_weights(regimes, ids=np.array(ids))
        w0 = {"id-1": 1.0}
        self.assertEqual(w["id-0"], w0)
        w = util.block_weights(regimes, ids=ids)
        w0 = {"id-1": 1.0}
        self.assertEqual(w["id-0"], w0)

    def test_comb(self):
        x = list(range(4))
        l = []
        for i in util.comb(x, 2):
            l.append(i)
        lo = [[0, 1], [0, 2], [0, 3], [1, 2], [1, 3], [2, 3]]
        self.assertEqual(l, lo)

    def test_order(self):
        w3 = util.order(self.w, kmax=3)
        w3105 = [1, -1, 1, 2, 1]
        self.assertEqual(w3105, w3[1][0:5])

    def test_higher_order(self):
        w10 = lat2W(10, 10)
        w10_2 = util.higher_order(w10, 2)
        w10_20 = {2: 1.0, 11: 1.0, 20: 1.0}
        self.assertEqual(w10_20, w10_2[0])
        w5 = lat2W()
        w50 = {1: 1.0, 5: 1.0}
        self.assertEqual(w50, w5[0])
        w51 = {0: 1.0, 2: 1.0, 6: 1.0}
        self.assertEqual(w51, w5[1])
        w5_2 = util.higher_order(w5, 2)
        w5_20 = {2: 1.0, 10: 1.0, 6: 1.0}
        self.assertEqual(w5_20, w5_2[0])

    def test_higher_order_sp(self):
        w10 = lat2W(10, 10)
        w10_3 = util.higher_order_sp(w10, 3)
        w10_30 = {30: 1.0, 21: 1.0, 12: 1.0, 3: 1.0}
        self.assertEqual(w10_30, w10_3[0])
        w10_3 = util.higher_order_sp(w10, 3, lower_order=True)
        w10_30 = {
            20: 1.0,
            30: 1.0,
            21: 1.0,
            10: 1.0,
            1: 1.0,
            11: 1.0,
            2: 1.0,
            12: 1.0,
            3: 1.0,
        }
        self.assertEqual(w10_30, w10_3[0])

    def test_higher_order_classes(self):
        wdb = DistanceBand.from_shapefile(examples.get_path("baltim.shp"), 34)
        wknn = KNN.from_shapefile(examples.get_path("baltim.shp"), 10)
        wrook = Rook.from_shapefile(examples.get_path("columbus.shp"))
        wqueen = Queen.from_shapefile(examples.get_path("columbus.shp"))
        wsparse = wqueen.sparse
        ww = W(wknn.neighbors, wknn.weights)
        util.higher_order(wdb, 2)
        util.higher_order(wknn, 3)
        util.higher_order(wrook, 4)
        util.higher_order(wqueen, 5)
        util.higher_order(wsparse, 2)
        util.higher_order(ww, 2)
        ww.transform = "r"
        wsparse_notbinary = wrook.sparse
        with self.assertRaises(ValueError):
            util.higher_order(wsparse, 2)
            util.higher_order(ww, 3)

    def test_shimbel(self):
        w5 = lat2W()
        w5_shimbel = util.shimbel(w5)
        w5_shimbel024 = 8
        self.assertEqual(w5_shimbel024, w5_shimbel[0][24])
        w5_shimbel004 = [-1, 1, 2, 3]
        self.assertEqual(w5_shimbel004, w5_shimbel[0][0:4])

    def test_full(self):
        neighbors = {
            "first": ["second"],
            "second": ["first", "third"],
            "third": ["second"],
        }
        weights = {"first": [1], "second": [1, 1], "third": [1]}
        w = W(neighbors, weights)
        wf, ids = util.full(w)
        wfo = np.array([[0.0, 1.0, 0.0], [1.0, 0.0, 1.0], [0.0, 1.0, 0.0]])
        np.testing.assert_array_almost_equal(wfo, wf, decimal=8)
        idso = ["first", "second", "third"]
        self.assertEqual(idso, ids)

    def test_full2W(self):
        a = np.zeros((4, 4))
        for i in range(len(a)):
            for j in range(len(a[i])):
                if i != j:
                    a[i, j] = np.random.random(1)
        w = util.full2W(a)
        np.testing.assert_array_equal(w.full()[0], a)
        ids = ["myID0", "myID1", "myID2", "myID3"]
        w = util.full2W(a, ids=ids)
        np.testing.assert_array_equal(w.full()[0], a)
        w.full()[0] == a

    def test_WSP2W(self):
        sp = util.lat2SW(2, 5)
        wsp = WSP(sp)
        w = util.WSP2W(wsp)
        self.assertEqual(w.n, 10)
        self.assertEqual(w[0], {1: 1, 5: 1})
        w = psopen(examples.get_path("sids2.gal"), "r").read()
        wsp = WSP(w.sparse, w.id_order)
        w = util.WSP2W(wsp)
        self.assertEqual(w.n, 100)
        self.assertEqual(
            w["37135"],
            {"37001": 1.0, "37033": 1.0, "37037": 1.0, "37063": 1.0, "37145": 1.0},
        )

    def test_fill_diagonal(self):
        w1 = util.fill_diagonal(self.w)
        r1 = {0: 1.0, 1: 1.0, 4: 1.0, 101: 1.0, 85: 1.0, 5: 1.0}
        self.assertEqual(w1[0], r1)
        w1 = util.fill_diagonal(self.w, 20)
        r1 = {0: 20, 1: 1.0, 4: 1.0, 101: 1.0, 85: 1.0, 5: 1.0}
        self.assertEqual(w1[0], r1)
        diag = np.arange(100, 100 + self.w.n)
        w1 = util.fill_diagonal(self.w, diag)
        r1 = {0: 100, 1: 1.0, 4: 1.0, 101: 1.0, 85: 1.0, 5: 1.0}
        self.assertEqual(w1[0], r1)

    def test_remap_ids(self):
        w = lat2W(3, 2)
        wid_order = [0, 1, 2, 3, 4, 5]
        self.assertEqual(wid_order, w.id_order)
        wneighbors0 = [2, 1]
        self.assertEqual(wneighbors0, w.neighbors[0])
        old_to_new = {0: "a", 1: "b", 2: "c", 3: "d", 4: "e", 5: "f"}
        w_new = util.remap_ids(w, old_to_new)
        w_newid_order = ["a", "b", "c", "d", "e", "f"]
        self.assertEqual(w_newid_order, w_new.id_order)
        w_newdneighborsa = ["c", "b"]
        self.assertEqual(w_newdneighborsa, w_new.neighbors["a"])

    def test_get_ids_shp(self):
        polyids = util.get_ids(examples.get_path("columbus.shp"), "POLYID")
        polyids5 = [1, 2, 3, 4, 5]
        self.assertEqual(polyids5, polyids[:5])

    @unittest.skipIf(
        not HAS_GEOPANDAS, "Missing geopandas, cannot test get_ids with gdf"
    )
    def test_get_ids_gdf(self):
        gdf = gpd.read_file(examples.get_path("columbus.shp"))
        polyids = util.get_ids(gdf, "POLYID")
        polyids5 = [1, 2, 3, 4, 5]
        self.assertEqual(polyids5, polyids[:5])

    def test_get_points_array_from_shapefile(self):
        xy = util.get_points_array_from_shapefile(examples.get_path("juvenile.shp"))
        xy3 = np.array([[94.0, 93.0], [80.0, 95.0], [79.0, 90.0]])
        np.testing.assert_array_almost_equal(xy3, xy[:3], decimal=8)
        xy = util.get_points_array_from_shapefile(examples.get_path("columbus.shp"))
        xy3 = np.array(
            [
                [8.82721847, 14.36907602],
                [8.33265837, 14.03162401],
                [9.01226541, 13.81971908],
            ]
        )
        np.testing.assert_array_almost_equal(xy3, xy[:3], decimal=8)

    def test_min_threshold_distance(self):
        x, y = np.indices((5, 5))
        x.shape = (25, 1)
        y.shape = (25, 1)
        data = np.hstack([x, y])
        mint = 1.0
        self.assertEqual(mint, util.min_threshold_distance(data))

    def test_attach_islands(self):
        w = Rook.from_shapefile(examples.get_path("10740.shp"))
        w_knn1 = KNN.from_shapefile(examples.get_path("10740.shp"), k=1)
        w_attach = util.attach_islands(w, w_knn1)
        self.assertEqual(w_attach.islands, [])
        self.assertEqual(w_attach[w.islands[0]], {166: 1.0})

    @unittest.skipIf(
        not HAS_GEOPANDAS, "Missing geopandas, cannot test nonplanar neighbors"
    )
    def test_nonplanar_neighbors(self):
        df = gpd.read_file(examples.get_path("map_RS_BR.shp"))
        w = Queen.from_dataframe(df)
        self.assertEqual(
            w.islands,
            [
                0,
                4,
                23,
                27,
                80,
                94,
                101,
                107,
                109,
                119,
                122,
                139,
                169,
                175,
                223,
                239,
                247,
                253,
                254,
                255,
                256,
                261,
                276,
                291,
                294,
                303,
                321,
                357,
                374,
            ],
        )
        wnp = nonplanar_neighbors(w, df)
        self.assertEqual(wnp.islands, [])
        self.assertEqual(w.neighbors[0], [])
        self.assertEqual(wnp.neighbors[0], [23, 59, 152, 239])
        self.assertEqual(wnp.neighbors[23], [0, 45, 59, 107, 152, 185, 246])

    @unittest.skipIf(
        not HAS_GEOPANDAS, "Missing geopandas, cannot test fuzzy_contiguity"
    )
    def test_fuzzy_contiguity(self):
        rs = examples.get_path("map_RS_BR.shp")
        rs_df = gpd.read_file(rs)
        wf = fuzzy_contiguity(rs_df)
        self.assertEqual(wf.islands, [])
        self.assertEqual(set(wf.neighbors[0]), set([239, 59, 152, 23]))
        buff = fuzzy_contiguity(rs_df, buffering=True, buffer=0.2)
        self.assertEqual(
            set(buff.neighbors[0]), set([175, 119, 239, 59, 152, 246, 23, 107])
        )
        rs_index = rs_df.set_index("NM_MUNICIP")
        index_w = fuzzy_contiguity(rs_index)
        self.assertEqual(
            set(index_w.neighbors["TAVARES"]), set(["SÃO JOSÉ DO NORTE", "MOSTARDAS"])
        )
        wf_pred = fuzzy_contiguity(rs_df, predicate="touches")
        self.assertEqual(set(wf_pred.neighbors[0]), set([]))
        self.assertEqual(set(wf_pred.neighbors[1]), set([142, 82, 197, 285, 386, 350]))


suite = unittest.TestLoader().loadTestsFromTestCase(Testutil)

if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite)



import unittest
from ..mat import MatIO
from .... import examples as pysal_examples
from ...fileio import FileIO as psopen
import tempfile
import os
import warnings


class test_MatIO(unittest.TestCase):
    def setUp(self):
        self.test_file = test_file = pysal_examples.get_path("spat-sym-us.mat")
        self.obj = MatIO(test_file, "r")

    def test_close(self):
        f = self.obj
        f.close()
        self.assertRaises(ValueError, f.read)

    def test_read(self):
        w = self.obj.read()
        self.assertEqual(46, w.n)
        self.assertEqual(4.0869565217391308, w.mean_neighbors)
        self.assertEqual([1.0, 1.0, 1.0, 1.0], list(w[1].values()))

    def test_seek(self):
        self.test_read()
        self.assertRaises(StopIteration, self.obj.read)
        self.obj.seek(0)
        self.test_read()

    def test_write(self):
        w = self.obj.read()
        f = tempfile.NamedTemporaryFile(suffix=".mat")
        fname = f.name
        f.close()
        o = psopen(fname, "w")
        with warnings.catch_warnings(record=True) as warn:
            warnings.simplefilter("always")
            o.write(w)
            if len(warn) > 0:
                assert issubclass(warn[0].category, FutureWarning)
        o.close()
        wnew = psopen(fname, "r").read()
        self.assertEqual(wnew.pct_nonzero, w.pct_nonzero)
        os.remove(fname)


if __name__ == "__main__":
    unittest.main()



"""Unit tests for gal.py"""
import unittest
import tempfile
import os
from ..gal import GalIO
from ...fileio import FileIO as psopen
from .... import examples as pysal_examples


class test_GalIO(unittest.TestCase):
    def setUp(self):
        self.test_file = test_file = pysal_examples.get_path("sids2.gal")
        self.obj = GalIO(test_file, "r")

    def test___init__(self):
        self.assertEqual(self.obj._typ, str)

    def test_close(self):
        f = self.obj
        f.close()
        self.assertRaises(ValueError, f.read)

    def test_read(self):
        # reading a GAL returns a W
        w = self.obj.read()
        self.assertEqual(w.n, 100)
        self.assertAlmostEqual(w.sd, 1.5151237573214935)
        self.assertEqual(w.s0, 462.0)
        self.assertEqual(w.s1, 924.0)

    def test_seek(self):
        self.test_read()
        self.assertRaises(StopIteration, self.obj.read)
        self.obj.seek(0)
        self.test_read()

    def test_write(self):
        w = self.obj.read()
        f = tempfile.NamedTemporaryFile(suffix=".gal")
        fname = f.name
        f.close()
        o = psopen(fname, "w")
        o.write(w)
        o.close()
        wnew = psopen(fname, "r").read()
        self.assertEqual(wnew.pct_nonzero, w.pct_nonzero)


if __name__ == "__main__":
    unittest.main()



from ... import cg
import re

__author__ = "Charles R Schmidt <schmidtc@gmail.com>"
__all__ = ["WKTParser"]


class WKTParser:
    """Class to represent `OGC WKT`, supports reading and writing.
    Modified from...
    # URL: http://dev.openlayers.org/releases/OpenLayers-2.7/lib/OpenLayers/Format/WKT.js
    # Reg Ex Strings copied from OpenLayers.Format.WKT

    Examples
    --------
    
    >>> import libpysal

    Create some Well-Known Text objects.

    >>> p = 'POLYGON((1 1, 5 1, 5 5, 1 5, 1 1), (2 2, 3 2, 3 3, 2 3, 2 2))'
    >>> pt = 'POINT(6 10)'
    >>> l = 'LINESTRING(3 4, 10 50, 20 25)'

    Instantiate the parser.

    >>> parser = libpysal.io.wkt.WKTParser()

    Inspect our WKT polygon.

    >>> parser(p).parts
    [[(1.0, 1.0), (1.0, 5.0), (5.0, 5.0), (5.0, 1.0), (1.0, 1.0)],
     [(2.0, 2.0), (2.0, 3.0), (3.0, 3.0), (3.0, 2.0), (2.0, 2.0)]]
    
    >>> parser(p).centroid
    (2.9705882352941178, 2.9705882352941178)
    
    >>> parser(p).area
    17.0

    Inspect ``pt``, our WKT point object.

    >>> parser(pt)
    (6.0, 10.0)

    Inspect our WKT linestring.

    >>> parser(l).len
    73.45538453219989
    
    >>> parser(l).parts
    [[(3.0, 4.0), (10.0, 50.0), (20.0, 25.0)]]

    Read in WKT from a file.

    >>> f = libpysal.io.open(libpysal.examples.get_path('stl_hom.wkt'))
    >>> f.mode
    'r'
    
    >>> f.header
    []

    See local doctest output for the items not tested.

    """

    regExes = {
        "typeStr": re.compile(r"^\s*([\w\s]+)\s*\(\s*(.*)\s*\)\s*$"),
        "spaces": re.compile(r"\s+"),
        "parenComma": re.compile(r"\)\s*,\s*\("),
        "doubleParenComma": re.compile(r"\)\s*\)\s*,\s*\(\s*\("),  # can't use {2} here
        "trimParens": re.compile(r"^\s*\(?(.*?)\)?\s*$"),
    }

    def __init__(self):
        self.parsers = p = {}
        p["point"] = self.Point
        p["linestring"] = self.LineString
        p["polygon"] = self.Polygon

    def Point(self, geoStr):
        """Returns a ``libpysal.cg.Point`` object."""
        coords = self.regExes["spaces"].split(geoStr.strip())
        return cg.Point((coords[0], coords[1]))

    def LineString(self, geoStr):
        """Returns a ``libpysal.cg.Chain`` object."""
        points = geoStr.strip().split(",")
        points = list(map(self.Point, points))
        return cg.Chain(points)

    def Polygon(self, geoStr):
        """Returns a ``libpysal.cg.Polygon`` object."""
        rings = self.regExes["parenComma"].split(geoStr.strip())
        for i, ring in enumerate(rings):
            ring = self.regExes["trimParens"].match(ring).groups()[0]
            ring = self.LineString(ring).vertices
            rings[i] = ring
        return cg.Polygon(rings)

    def fromWKT(self, wkt):
        """Returns geometric representation from WKT or ``None``.
        
        Raises
        ------
        NotImplementedError
            Raised when a unknown/unsupported format is passed in.
        
        """

        matches = self.regExes["typeStr"].match(wkt)
        if matches:
            geoType, geoStr = matches.groups()
            geoType = geoType.lower().strip()
            try:
                return self.parsers[geoType](geoStr)
            except KeyError:
                raise NotImplementedError("Unsupported WKT Type: %s." % geoType)
        else:
            return None

    __call__ = fromWKT


if __name__ == "__main__":

    p = "POLYGON((1 1,5 1,5 5,1 5,1 1),(2 2, 3 2, 3 3, 2 3,2 2))"
    pt = "POINT(6 10)"
    l = "LINESTRING(3 4,10 50,20 25)"
    wktExamples = [
        "POINT(6 10)",
        "LINESTRING(3 4,10 50,20 25)",
        "POLYGON((1 1,5 1,5 5,1 5,1 1),(2 2, 3 2, 3 3, 2 3,2 2))",
        "MULTIPOINT(3.5 5.6,4.8 10.5)",
        "MULTILINESTRING((3 4,10 50,20 25),(-5 -8,-10 -8,-15 -4))",
        "MULTIPOLYGON(((1 1,5 1,5 5,1 5,1 1),(2 2, 3 2, 3 3, 2 3,2 2)),((3 3,6 2,6 4,3 3)))",
        "GEOMETRYCOLLECTION(POINT(4 6),LINESTRING(4 6,7 10))",
        "POINT ZM (1 1 5 60)",
        "POINT M (1 1 80)",
        "POINT EMPTY",
        "MULTIPOLYGON EMPTY",
    ]
    wkt = WKTParser()



__all__ = ["DataTable"]

from . import fileio
from ..common import requires
from warnings import warn
import numpy as np

__author__ = "Charles R Schmidt <schmidtc@gmail.com>"


class DataTable(fileio.FileIO):
    """`DataTable` provides additional functionality to `FileIO`
    for data table file tables `FileIO` handlers that provide data
    tables should subclass this instead of `FileIO`.
    """

    class _By_Col:
        def __init__(self, parent):

            self.p = parent

        def __repr__(self) -> str:

            return "keys: " + self.p.header.__repr__()

        def __getitem__(self, key):

            return self.p._get_col(key)

        def __setitem__(self, key, val):

            self.p.cast(key, val)

        def __call__(self, key):

            return self.p._get_col(key)

    def __init__(self, *args, **kwargs):

        fileio.FileIO.__init__(self, *args, **kwargs)

    def __repr__(self) -> str:

        return "DataTable: %s" % self.dataPath

    def __len__(self):
        """__len__ should be implemented by `DataTable` subclasses."""

        raise NotImplementedError

    @property
    def by_col(self):
        return self._By_Col(self)

    def _get_col(self, key):
        """Returns the column vector.
        
        Raises
        ------
        AttributeError
            Raised when the header is not set.
        AttributeError
            Raised when a field does not exist.
            
        """

        if not self.header:
            raise AttributeError("Please set the header.")

        if key in self.header:
            return self[:, self.header.index(key)]
        else:
            raise AttributeError("Field: %s does not exist in header." % key)

    def by_col_array(self, *args):
        """Return columns of table as a ``numpy.ndarray``.

        Parameters
        ----------
        *args : iterable
            Any number of strings of length :math:`k` names of variables to extract.

        Returns
        -------
        results : numpy.ndarray
            An array of shape :math:`(n,k)`.

        Notes
        -----

        If the variables are not all of the same data type, then ``numpy`` rules
        for casting will result in a uniform type applied to all variables. If only
        strings are passed to the function, then an array with those columns will be
        constructed. If only one list of strings is passed, the output is identical
        to those strings being passed. If at least one list is passed and other
        strings or lists are passed, this returns a tuple containing arrays
        constructed from each positional argument.

        Examples
        --------

        >>> import libpysal
        >>> dbf = libpysal.io.open(libpysal.examples.get_path('NAT.dbf'))
        >>> hr = dbf.by_col_array('HR70', 'HR80')
        >>> hr[0:5]
        array([[ 0.        ,  8.85582713],
               [ 0.        , 17.20874204],
               [ 1.91515848,  3.4507747 ],
               [ 1.28864319,  3.26381409],
               [ 0.        ,  7.77000777]])
        
        >>> hr = dbf.by_col_array(['HR80', 'HR70'])
        >>> hr[0:5]
        array([[ 8.85582713,  0.        ],
               [17.20874204,  0.        ],
               [ 3.4507747 ,  1.91515848],
               [ 3.26381409,  1.28864319],
               [ 7.77000777,  0.        ]])
        
        >>> hr = dbf.by_col_array(['HR80'])
        >>> hr[0:5]
        array([[ 8.85582713],
               [17.20874204],
               [ 3.4507747 ],
               [ 3.26381409],
               [ 7.77000777]])
        
        Numpy only supports homogeneous arrays. See Notes above.

        >>> hr = dbf.by_col_array('STATE_NAME', 'HR80')
        >>> hr[0:5]
        array([['Minnesota', '8.8558271343'],
               ['Washington', '17.208742041'],
               ['Washington', '3.4507746989'],
               ['Washington', '3.2638140931'],
               ['Washington', '7.77000777']], dtype='<U20')

        >>> y, X = dbf.by_col_array('STATE_NAME', ['HR80', 'HR70'])
        >>> y[0:5]
        array([['Minnesota'],
               ['Washington'],
               ['Washington'],
               ['Washington'],
               ['Washington']], dtype='<U20')
        
        >>> X[0:5]
        array([[ 8.85582713,  0.        ],
               [17.20874204,  0.        ],
               [ 3.4507747 ,  1.91515848],
               [ 3.26381409,  1.28864319],
               [ 7.77000777,  0.        ]])
        """

        if any([isinstance(arg, list) for arg in args]):
            results = []
            for namelist in args:
                if isinstance(namelist, str):
                    results.append([self._get_col(namelist)])
                else:
                    results.append([self._get_col(vbl) for vbl in namelist])
            if len(results) == 1:
                results = np.array(results[0]).T
            else:
                results = tuple(np.array(lst).T for lst in results)
        else:
            results = np.array([self._get_col(name) for name in args]).T

        return results

    def __getitem__(self, key) -> list:
        """DataTables fully support slicing in 2D. To provide slicing, handlers
        must provide ``__len__``. Slicing accepts up to two arguments. For example,
        
        * ``table[row]``
        * ``table[row, col]``
        * ``table[row_start:row_stop]``
        * ``table[row_start:row_stop:row_step]``
        * ``table[:, col]``
        * ``table[:, col_start:col_stop]``
        * etc.

        ALL indices are Zero-Offsets. For example,
        
        * ``>>> assert index in range(0, len(table))``
        
        Raises
        ------
        TypeError
            Raised when two dimensions are not provided for slicing.
        TypeError
            Raised when an unknown key is present.
        
        """

        prevPos = self.tell()

        if issubclass(type(key), str):
            raise TypeError("index should be int or slice")

        if issubclass(type(key), int) or isinstance(key, slice):
            rows = key
            cols = None
        elif len(key) > 2:
            raise TypeError(
                "DataTables support two dimmensional slicing, % d slices provided."
                % len(key)
            )
        elif len(key) == 2:
            rows, cols = key
        else:
            raise TypeError("Key: % r, is confusing me. I don't know what to do." % key)

        if isinstance(rows, slice):
            row_start, row_stop, row_step = rows.indices(len(self))
            self.seek(row_start)
            data = [next(self) for i in range(row_start, row_stop, row_step)]
        else:
            self.seek(slice(rows).indices(len(self))[1])
            data = [next(self)]

        if cols is not None:
            if isinstance(cols, slice):
                col_start, col_stop, col_step = cols.indices(len(data[0]))
                data = [r[col_start:col_stop:col_step] for r in data]
            else:
                # col_start, col_stop, col_step = cols, cols+1, 1
                data = [r[cols] for r in data]
        self.seek(prevPos)

        return data

    @requires("pandas")
    def to_df(self, n=-1, read_shp=False, **df_kws):
        """Convert a ``libpysal.DataTable`` to a ``pandas.DataFrame``.
        
        Parameters
        ----------
        n : int
            Lines to read from file. Default is ``-1``.
        read_shp : bool
            Read in from a shapefile (``True``). Default is ``False``.
        **df_kws : dict
            Optional keyword arguments to pass into ``pandas.DataFrame()``.
        
        Returns
        -------
        df : pandas.DataFrame
            Pandas dataframe representation of the data.

        """

        import pandas as pd

        self.seek(0)
        header = self.header
        records = self.read(n)
        df = pd.DataFrame(records, columns=header, **df_kws)
        if read_shp is not False:
            if read_shp is True or self.dataPath.endswith(".dbf"):
                read_shp = self.dataPath[:-3] + "shp"
            try:
                from .geotable.shp import shp2series

                df["geometry"] = shp2series(self.dataPath[:-3] + "shp")
            except IOError as e:
                warn(
                    "Encountered the following error in attempting to read"
                    " the shapefile {}. Proceeding with read, but the error"
                    " will be reproduced below:\n"
                    " {}".format(self.dataPath[:-3] + "shp", e)
                )
        return df


def _test():
    import doctest

    doctest.testmod(verbose=True)


if __name__ == "__main__":
    _test()



import unittest
from .... import examples as pysal_examples
from .. import csvWrapper
from ...util import WKTParser
import tempfile
import os
from sys import version as V

PY3 = int(V[0]) > 2


class test_csvWrapper(unittest.TestCase):
    def setUp(self):
        stl = pysal_examples.load_example("stl")
        self.test_file = test_file = stl.get_path("stl_hom.csv")
        self.obj = csvWrapper.csvWrapper(test_file, "r")

    def test_len(self):
        self.assertEqual(len(self.obj), 78)

    def test_tell(self):
        self.assertEqual(self.obj.tell(), 0)
        self.obj.read(1)
        self.assertEqual(self.obj.tell(), 1)
        self.obj.read(50)
        self.assertEqual(self.obj.tell(), 51)
        self.obj.read()
        self.assertEqual(self.obj.tell(), 78)

    def test_seek(self):
        self.obj.seek(0)
        self.assertEqual(self.obj.tell(), 0)
        self.obj.seek(55)
        self.assertEqual(self.obj.tell(), 55)
        self.obj.read(1)
        self.assertEqual(self.obj.tell(), 56)

    def test_read(self):
        self.obj.seek(0)
        objs = self.obj.read()
        self.assertEqual(len(objs), 78)
        self.obj.seek(0)
        objsB = list(self.obj)
        self.assertEqual(len(objsB), 78)
        for rowA, rowB in zip(objs, objsB):
            self.assertEqual(rowA, rowB)

    def test_casting(self):
        self.obj.cast("WKT", WKTParser())
        verts = [
            (-89.585220336914062, 39.978794097900391),
            (-89.581146240234375, 40.094867706298828),
            (-89.603988647460938, 40.095306396484375),
            (-89.60589599609375, 40.136119842529297),
            (-89.6103515625, 40.3251953125),
            (-89.269027709960938, 40.329566955566406),
            (-89.268562316894531, 40.285579681396484),
            (-89.154655456542969, 40.285774230957031),
            (-89.152763366699219, 40.054969787597656),
            (-89.151618957519531, 39.919403076171875),
            (-89.224777221679688, 39.918678283691406),
            (-89.411857604980469, 39.918041229248047),
            (-89.412437438964844, 39.931644439697266),
            (-89.495201110839844, 39.933486938476562),
            (-89.4927978515625, 39.980186462402344),
            (-89.585220336914062, 39.978794097900391),
        ]
        if PY3:
            for i, pt in enumerate(self.obj.__next__()[0].vertices):
                self.assertEqual(pt[:], verts[i])
        else:
            for i, pt in enumerate(self.obj.next()[0].vertices):
                self.assertEqual(pt[:], verts[i])

    def test_by_col(self):
        for field in self.obj.header:
            self.assertEqual(len(self.obj.by_col[field]), 78)

    def test_slicing(self):
        chunk = self.obj[50:55, 1:3]
        self.assertEqual(chunk[0], ["Jefferson", "Missouri"])
        self.assertEqual(chunk[1], ["Jefferson", "Illinois"])
        self.assertEqual(chunk[2], ["Miller", "Missouri"])
        self.assertEqual(chunk[3], ["Maries", "Missouri"])
        self.assertEqual(chunk[4], ["White", "Illinois"])


if __name__ == "__main__":
    unittest.main()



from ...weights.contiguity import Rook, Queen
from ..fileio import FileIO as ps_open
from .utils import insert_metadata
import os
from .shp import shp2series, series2shp
from .dbf import dbf2df, df2dbf


def read_files(filepath, **kwargs):
    """Reads a ``.dbf``/``.shp`` pair, squashing geometries into a 'geometry' column.
    
    Parameters
    ----------
    filepath : str
        The file path.
    **kwargs : dict
        Optional keyword arguments for ``dbf2df()``.
    
    Returns
    -------
    df : pandas.DataFrame
        The results dataframe returned from ``dbf2df()``.
    
    """

    # keyword arguments wrapper will strip all around dbf2df's required arguments
    geomcol = kwargs.pop("geomcol", "geometry")
    weights = kwargs.pop("weights", "")

    dbf_path, shp_path = _pairpath(filepath)

    df = dbf2df(dbf_path, **kwargs)
    df[geomcol] = shp2series(shp_path)

    if weights != "" and isinstance(weights, str):
        if weights.lower() in ["rook", "queen"]:
            if weights.lower() == "rook":
                W = Rook.from_dataframe(df, geometr)
            else:
                W = Queen.from_dataframe(df)
            insert_metadata(df, W, name="W", inplace=True)
        else:
            try:
                W_path = os.path.splitext(dbf_path)[0] + "." + weights
                W = ps_open(W_path).read()
                insert_metadata(df, W, name="W", inplace=True)
            except IOError:
                print("Weights construction failed! Passing on weights.")

    return df


def write_files(df, filepath, **kwargs):
    """Writes dataframes with potential geometric components out to files.
    
    Parameters
    ----------
    df : pandas.DataFrame
        The dataframe to write out.
    filepath : str
        The file path.
    **kwargs : dict
        Optional keyword arguments for ``df2dbf()``.
    
    Returns
    -------
    dbf_path : str
        Path to the output ``.dbf``
    paths : tuple
        The file paths for ``dbf_out``, ``shp_out``, ``W_path``.
    
    """

    geomcol = kwargs.pop("geomcol", "geometry")
    weights = kwargs.pop("weights", "gal")

    dbf_path, shp_path = _pairpath(filepath)

    if geomcol not in df.columns:
        dbf_path = df2dbf(df, dbf_path, **kwargs)
        return dbf_path
    else:
        shp_out = series2shp(df[geomcol], shp_path)
        not_geom = [x for x in df.columns if x != geomcol]
        dbf_out = df2dbf(df[not_geom], dbf_path, **kwargs)

        if hasattr(df, "W"):
            W_path = os.path.splitext(filepath)[0] + "." + weights
            ps_open(W_path, "w").write(df.W)
        else:
            W_path = "no weights written"

        paths = dbf_out, shp_out, W_path

        return paths


def _pairpath(filepath: str) -> tuple:
    """Return ``.dbf``/``.shp`` paths for any ``.shp``,
    ``.dbf``, or basepath passed to function. 
    
    """

    base = os.path.splitext(filepath)[0]
    paths = base + ".dbf", base + ".shp"

    return paths



"""The :mod:`libpysal.examples` module provides example datasets.

The datasets consist of two sets, built-ins which are installed with
this module and remotes that can be downloaded.

This module provides functionality for working with these example datasets.
"""


import pandas as pd
from .base import example_manager
from .remotes import datasets as remote_datasets
from .builtin import datasets as builtin_datasets
from typing import Union


available_datasets = builtin_datasets.copy()
available_datasets.update(remote_datasets.datasets)

__all__ = ["get_path", "available", "explain", "fetch_all",
           "get_url", "load_example", "summary"]

example_manager.add_examples(available_datasets)


def fetch_all():
    """Fetch and install all remote datasets."""
    datasets = remote_datasets.datasets
    names = list(datasets.keys())
    names.sort()
    for name in names:
        example = datasets[name]
        try:
            example.download()
        except:
            print("Example not downloaded: {}".format(name))
    example_manager.add_examples(datasets)


def available() -> pd.DataFrame:
    """Return a dataframe with available datasets."""
    return example_manager.available()


def explain(name: str) -> str:
    """Explain a dataset by name."""
    return example_manager.explain(name)


def get_url(name: str) -> str:
    """Get url for remote dataset."""
    return example_manager.get_remote_url(name)


def load_example(example_name: str) -> Union[base.Example, builtin.LocalExample]:
    """Load example dataset instance."""
    example = example_manager.load(example_name)
    return example


def get_path(file_name: str) -> str:
    """Get the path for a file by searching installed datasets."""
    installed = example_manager.get_installed_names()
    for name in installed:
        example = example_manager.datasets[name]
        pth = example.get_path(file_name, verbose=False)
        if pth:
            return pth
    print("{} is not a file in any installed dataset.".format(file_name))


def summary():
    """Summary of datasets."""
    example_manager.summary()



from ...cg.shapes import asShape as pShape
from ...common import requires as _requires
from warnings import warn


@_requires("geopandas")
def to_df(df, geom_col="geometry", **kw):
    """Convert a ``geopandas.GeoDataFrame`` into a normal
    ``pandas.DataFrame`` with a column containing PySAL shapes.

    Parameters
    ----------
    df : geopandas.GeoDataFrame
        A ``geopandas.GeoDataFrame`` (or ``pandas.DataFrame``)
        with a column containing geo-interfaced shapes.
    geom_col : str
        The column name in ``df`` contains the geometry. Default is ``'geometry'``.
    **kw : dict
        Optional keyword arguments for ``pandas.DataFrame()``.

    Returns
    -------
    df : pandas.DataFrame
        The data converted into a ``pandas.DataFrame`` object.

    See Also
    --------

    pandas.DataFrame

    """

    import pandas as pd
    from geopandas import GeoDataFrame, GeoSeries

    df[geom_col] = df[geom_col].apply(pShape)

    if isinstance(df, (GeoDataFrame, GeoSeries)):
        df = pd.DataFrame(df, **kw)

    return df


@_requires("geopandas")
def to_gdf(df, geom_col="geometry", **kw):
    """Convert a ``pandas.DataFrame`` with
    geometry column to a ``geopandas.GeoDataFrame``.

    Parameters
    ----------
    df : pandas.DataFrame
        A ``pandas.DataFrame`` with a column containing geo-interfaced shapes.
    geom_col : str
        The column name in ``df`` contains the geometry. Default is ``'geometry'``.
    **kw : dict
        Optional keyword arguments for ``geopandas.GeoDataFrame()``.

    Returns
    -------
    gdf : geopandas.GeoDataFrame
        The data converted into a ``geopandas.GeoDataFrame`` object.

    See Also
    --------

    geopandas.GeoDataFrame

    """

    from geopandas import GeoDataFrame
    from shapely.geometry import shape

    df[geom_col] = df[geom_col].apply(shape)

    gdf = GeoDataFrame(df, geometry=geom_col, **kw)

    return gdf


def insert_metadata(df, obj, name=None, inplace=True, overwrite=False):
    """Insert/update metadata for a dataframe."""

    if not inplace:
        new = df.copy(deep=True)
        insert_metadata(new, obj, name=name, inplace=True)
        return new

    if name is None:
        name = type(obj).__name__

    if hasattr(df, name):
        if overwrite:
            warn("Overwriting attribute {}! This may break the dataframe!".format(name))
        else:
            raise Exception(
                "Dataframe already has attribute {}. Cowardly refusing "
                "to break dataframe.".format(name)
            )

    df._metadata.append(name)
    df.__setattr__(name, obj)



import os
import tempfile

import unittest
from ..weights import W, WSP
from .. import util
from ..util import WSP2W, lat2W
from ..contiguity import Rook
from ...io.fileio import FileIO as psopen
from ... import examples
from ..distance import KNN
import numpy as np

NPTA3E = np.testing.assert_array_almost_equal


class TestW(unittest.TestCase):
    def setUp(self):
        self.w = Rook.from_shapefile(
            examples.get_path("10740.shp"), silence_warnings=True
        )

        self.neighbors = {
            0: [3, 1],
            1: [0, 4, 2],
            2: [1, 5],
            3: [0, 6, 4],
            4: [1, 3, 7, 5],
            5: [2, 4, 8],
            6: [3, 7],
            7: [4, 6, 8],
            8: [5, 7],
        }
        self.weights = {
            0: [1, 1],
            1: [1, 1, 1],
            2: [1, 1],
            3: [1, 1, 1],
            4: [1, 1, 1, 1],
            5: [1, 1, 1],
            6: [1, 1],
            7: [1, 1, 1],
            8: [1, 1],
        }

        self.w3x3 = util.lat2W(3, 3)

    def test_W(self):
        w = W(self.neighbors, self.weights, silence_warnings=True)
        self.assertEqual(w.pct_nonzero, 29.62962962962963)

    def test___getitem__(self):
        self.assertEqual(self.w[0], {1: 1.0, 4: 1.0, 101: 1.0, 85: 1.0, 5: 1.0})

    def test___init__(self):
        w = W(self.neighbors, self.weights, silence_warnings=True)
        self.assertEqual(w.pct_nonzero, 29.62962962962963)

    def test___iter__(self):
        w = lat2W(3, 3)
        res = {}
        for i, wi in enumerate(w):
            res[i] = wi
        self.assertEqual(res[0], (0, {1: 1.0, 3: 1.0}))
        self.assertEqual(res[8], (8, {5: 1.0, 7: 1.0}))

    def test_asymmetries(self):
        w = lat2W(3, 3)
        w.transform = "r"
        result = w.asymmetry()
        self.assertEqual(
            result,
            [
                (0, 1),
                (0, 3),
                (1, 0),
                (1, 2),
                (1, 4),
                (2, 1),
                (2, 5),
                (3, 0),
                (3, 4),
                (3, 6),
                (4, 1),
                (4, 3),
                (4, 5),
                (4, 7),
                (5, 2),
                (5, 4),
                (5, 8),
                (6, 3),
                (6, 7),
                (7, 4),
                (7, 6),
                (7, 8),
                (8, 5),
                (8, 7),
            ],
        )

    def test_asymmetry(self):
        w = lat2W(3, 3)
        self.assertEqual(w.asymmetry(), [])
        w.transform = "r"
        self.assertFalse(w.asymmetry() == [])

    def test_cardinalities(self):
        w = lat2W(3, 3)
        self.assertEqual(
            w.cardinalities, {0: 2, 1: 3, 2: 2, 3: 3, 4: 4, 5: 3, 6: 2, 7: 3, 8: 2}
        )

    def test_diagW2(self):
        NPTA3E(
            self.w3x3.diagW2, np.array([2.0, 3.0, 2.0, 3.0, 4.0, 3.0, 2.0, 3.0, 2.0])
        )

    def test_diagWtW(self):
        NPTA3E(
            self.w3x3.diagW2, np.array([2.0, 3.0, 2.0, 3.0, 4.0, 3.0, 2.0, 3.0, 2.0])
        )

    def test_diagWtW_WW(self):
        NPTA3E(
            self.w3x3.diagWtW_WW,
            np.array([4.0, 6.0, 4.0, 6.0, 8.0, 6.0, 4.0, 6.0, 4.0]),
        )

    def test_full(self):
        wf = np.array(
            [
                [0.0, 1.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [1.0, 0.0, 1.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 1.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0],
                [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.0],
            ]
        )
        ids = list(range(9))

        wf1, ids1 = self.w3x3.full()
        NPTA3E(wf1, wf)
        self.assertEqual(ids1, ids)

    def test_get_transform(self):
        self.assertEqual(self.w3x3.transform, "O")
        self.w3x3.transform = "r"
        self.assertEqual(self.w3x3.transform, "R")
        self.w3x3.transform = "b"

    def test_higher_order(self):
        weights = {
            0: [1.0, 1.0, 1.0],
            1: [1.0, 1.0, 1.0],
            2: [1.0, 1.0, 1.0],
            3: [1.0, 1.0, 1.0],
            4: [1.0, 1.0, 1.0, 1.0],
            5: [1.0, 1.0, 1.0],
            6: [1.0, 1.0, 1.0],
            7: [1.0, 1.0, 1.0],
            8: [1.0, 1.0, 1.0],
        }
        neighbors = {
            0: [4, 6, 2],
            1: [3, 5, 7],
            2: [8, 0, 4],
            3: [7, 1, 5],
            4: [8, 0, 2, 6],
            5: [1, 3, 7],
            6: [4, 0, 8],
            7: [3, 1, 5],
            8: [6, 2, 4],
        }
        wneighbs = {
            k: {neighb: weights[k][i] for i, neighb in enumerate(v)}
            for k, v in list(neighbors.items())
        }
        w2 = util.higher_order(self.w3x3, 2)
        test_wneighbs = {
            k: {ne: weights[k][i] for i, ne in enumerate(v)}
            for k, v in list(w2.neighbors.items())
        }
        self.assertEqual(test_wneighbs, wneighbs)

    def test_histogram(self):
        hist = [
            (0, 1),
            (1, 1),
            (2, 4),
            (3, 20),
            (4, 57),
            (5, 44),
            (6, 36),
            (7, 15),
            (8, 7),
            (9, 1),
            (10, 6),
            (11, 0),
            (12, 2),
            (13, 0),
            (14, 0),
            (15, 1),
        ]
        self.assertEqual(self.w.histogram, hist)

    def test_id2i(self):
        id2i = {0: 0, 1: 1, 2: 2, 3: 3, 4: 4, 5: 5, 6: 6, 7: 7, 8: 8}
        self.assertEqual(self.w3x3.id2i, id2i)

    def test_id_order_set(self):
        w = W(neighbors={"a": ["b"], "b": ["a", "c"], "c": ["b"]})
        self.assertFalse(w.id_order_set)

    def test_islands(self):
        w = W(
            neighbors={"a": ["b"], "b": ["a", "c"], "c": ["b"], "d": []},
            silence_warnings=True,
        )
        self.assertEqual(w.islands, ["d"])
        self.assertEqual(self.w3x3.islands, [])

    def test_max_neighbors(self):
        w = W(
            neighbors={"a": ["b"], "b": ["a", "c"], "c": ["b"], "d": []},
            silence_warnings=True,
        )
        self.assertEqual(w.max_neighbors, 2)
        self.assertEqual(self.w3x3.max_neighbors, 4)

    def test_mean_neighbors(self):
        w = util.lat2W()
        self.assertEqual(w.mean_neighbors, 3.2)

    def test_min_neighbors(self):
        w = util.lat2W()
        self.assertEqual(w.min_neighbors, 2)

    def test_n(self):
        w = util.lat2W()
        self.assertEqual(w.n, 25)

    def test_neighbor_offsets(self):
        d = {
            0: [3, 1],
            1: [0, 4, 2],
            2: [1, 5],
            3: [0, 6, 4],
            4: [1, 3, 7, 5],
            5: [2, 4, 8],
            6: [3, 7],
            7: [4, 6, 8],
            8: [5, 7],
        }

        self.assertEqual(self.w3x3.neighbor_offsets, d)

    def test_nonzero(self):
        self.assertEqual(self.w3x3.nonzero, 24)

    def test_order(self):
        w = util.lat2W(3, 3)
        o = {
            0: [-1, 1, 2, 1, 2, 3, 2, 3, 0],
            1: [1, -1, 1, 2, 1, 2, 3, 2, 3],
            2: [2, 1, -1, 3, 2, 1, 0, 3, 2],
            3: [1, 2, 3, -1, 1, 2, 1, 2, 3],
            4: [2, 1, 2, 1, -1, 1, 2, 1, 2],
            5: [3, 2, 1, 2, 1, -1, 3, 2, 1],
            6: [2, 3, 0, 1, 2, 3, -1, 1, 2],
            7: [3, 2, 3, 2, 1, 2, 1, -1, 1],
            8: [0, 3, 2, 3, 2, 1, 2, 1, -1],
        }
        self.assertEqual(util.order(w), o)

    def test_pct_nonzero(self):
        self.assertEqual(self.w3x3.pct_nonzero, 29.62962962962963)

    def test_s0(self):
        self.assertEqual(self.w3x3.s0, 24.0)

    def test_s1(self):
        self.assertEqual(self.w3x3.s1, 48.0)

    def test_s2(self):
        self.assertEqual(self.w3x3.s2, 272.0)

    def test_s2array(self):
        s2a = np.array(
            [[16.0], [36.0], [16.0], [36.0], [64.0], [36.0], [16.0], [36.0], [16.0]]
        )
        NPTA3E(self.w3x3.s2array, s2a)

    def test_sd(self):
        self.assertEqual(self.w3x3.sd, 0.66666666666666663)

    def test_set_transform(self):
        w = util.lat2W(2, 2)
        self.assertEqual(w.transform, "O")
        self.assertEqual(w.weights[0], [1.0, 1.0])
        w.transform = "r"
        self.assertEqual(w.weights[0], [0.5, 0.5])

    def test_shimbel(self):
        d = {
            0: [-1, 1, 2, 1, 2, 3, 2, 3, 4],
            1: [1, -1, 1, 2, 1, 2, 3, 2, 3],
            2: [2, 1, -1, 3, 2, 1, 4, 3, 2],
            3: [1, 2, 3, -1, 1, 2, 1, 2, 3],
            4: [2, 1, 2, 1, -1, 1, 2, 1, 2],
            5: [3, 2, 1, 2, 1, -1, 3, 2, 1],
            6: [2, 3, 4, 1, 2, 3, -1, 1, 2],
            7: [3, 2, 3, 2, 1, 2, 1, -1, 1],
            8: [4, 3, 2, 3, 2, 1, 2, 1, -1],
        }
        self.assertEqual(util.shimbel(self.w3x3), d)

    def test_sparse(self):
        self.assertEqual(self.w3x3.sparse.nnz, 24)

    def test_trcW2(self):
        self.assertEqual(self.w3x3.trcW2, 24.0)

    def test_trcWtW(self):
        self.assertEqual(self.w3x3.trcWtW, 24.0)

    def test_trcWtW_WW(self):
        self.assertEqual(self.w3x3.trcWtW_WW, 48.0)

    def test_symmetrize(self):
        symm = self.w.symmetrize()
        np.testing.assert_allclose(symm.sparse.toarray(), self.w.sparse.toarray())
        knn = KNN.from_shapefile(
            examples.get_path("baltim.shp"), k=10, silence_warnings=True
        )
        sknn = knn.symmetrize()
        assert not np.allclose(knn.sparse.toarray(), sknn.sparse.toarray())
        np.testing.assert_allclose(sknn.sparse.toarray(), sknn.sparse.toarray().T)
        knn.symmetrize(inplace=True)
        np.testing.assert_allclose(sknn.sparse.toarray(), knn.sparse.toarray())
        np.testing.assert_allclose(knn.sparse.toarray().T, knn.sparse.toarray())

    def test_connected_components(self):
        disco = {0: [1], 1: [0], 2: [3], 3: [2]}
        disco = W(disco)
        assert disco.n_components == 2

    def test_roundtrip_write(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            path = os.path.join(str(tmpdir), "tmp.gal")
            self.w.to_file(path)
            new = W.from_file(path)
        np.testing.assert_array_equal(self.w.sparse.toarray(), new.sparse.toarray())


class Test_WSP_Back_To_W(unittest.TestCase):
    # Test to make sure we get back to the same W functionality
    def setUp(self):
        self.w = Rook.from_shapefile(
            examples.get_path("10740.shp"), silence_warnings=True
        )
        wsp = self.w.to_WSP()
        self.w = wsp.to_W(silence_warnings=True)

        self.neighbors = {
            0: [3, 1],
            1: [0, 4, 2],
            2: [1, 5],
            3: [0, 6, 4],
            4: [1, 3, 7, 5],
            5: [2, 4, 8],
            6: [3, 7],
            7: [4, 6, 8],
            8: [5, 7],
        }
        self.weights = {
            0: [1, 1],
            1: [1, 1, 1],
            2: [1, 1],
            3: [1, 1, 1],
            4: [1, 1, 1, 1],
            5: [1, 1, 1],
            6: [1, 1],
            7: [1, 1, 1],
            8: [1, 1],
        }

        self.w3x3 = util.lat2W(3, 3)
        w3x3 = WSP(self.w3x3.sparse, self.w3x3.id_order)
        self.w3x3 = WSP2W(w3x3)

    def test_W(self):
        w = W(self.neighbors, self.weights, silence_warnings=True)
        self.assertEqual(w.pct_nonzero, 29.62962962962963)

    def test___getitem__(self):
        self.assertEqual(self.w[0], {1: 1.0, 4: 1.0, 101: 1.0, 85: 1.0, 5: 1.0})

    def test___init__(self):
        w = W(self.neighbors, self.weights, silence_warnings=True)
        self.assertEqual(w.pct_nonzero, 29.62962962962963)

    def test___iter__(self):
        w = util.lat2W(3, 3)
        res = {}
        for i, wi in enumerate(w):
            res[i] = wi
        self.assertEqual(res[0], (0, {1: 1.0, 3: 1.0}))
        self.assertEqual(res[8], (8, {5: 1.0, 7: 1.0}))

    def test_asymmetries(self):
        w = util.lat2W(3, 3)
        w.transform = "r"
        result = w.asymmetry()
        self.assertEqual(
            result,
            [
                (0, 1),
                (0, 3),
                (1, 0),
                (1, 2),
                (1, 4),
                (2, 1),
                (2, 5),
                (3, 0),
                (3, 4),
                (3, 6),
                (4, 1),
                (4, 3),
                (4, 5),
                (4, 7),
                (5, 2),
                (5, 4),
                (5, 8),
                (6, 3),
                (6, 7),
                (7, 4),
                (7, 6),
                (7, 8),
                (8, 5),
                (8, 7),
            ],
        )

    def test_asymmetry(self):
        w = util.lat2W(3, 3)
        self.assertEqual(w.asymmetry(), [])
        w.transform = "r"
        self.assertFalse(w.asymmetry() == [])

    def test_cardinalities(self):
        w = util.lat2W(3, 3)
        self.assertEqual(
            w.cardinalities, {0: 2, 1: 3, 2: 2, 3: 3, 4: 4, 5: 3, 6: 2, 7: 3, 8: 2}
        )

    def test_diagW2(self):
        NPTA3E(
            self.w3x3.diagW2, np.array([2.0, 3.0, 2.0, 3.0, 4.0, 3.0, 2.0, 3.0, 2.0])
        )

    def test_diagWtW(self):
        NPTA3E(
            self.w3x3.diagW2, np.array([2.0, 3.0, 2.0, 3.0, 4.0, 3.0, 2.0, 3.0, 2.0])
        )

    def test_diagWtW_WW(self):
        NPTA3E(
            self.w3x3.diagWtW_WW,
            np.array([4.0, 6.0, 4.0, 6.0, 8.0, 6.0, 4.0, 6.0, 4.0]),
        )

    def test_full(self):
        wf = np.array(
            [
                [0.0, 1.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [1.0, 0.0, 1.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 1.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0],
                [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.0],
            ]
        )
        ids = list(range(9))

        wf1, ids1 = self.w3x3.full()
        NPTA3E(wf1, wf)
        self.assertEqual(ids1, ids)

    def test_get_transform(self):
        self.assertEqual(self.w3x3.transform, "O")
        self.w3x3.transform = "r"
        self.assertEqual(self.w3x3.transform, "R")
        self.w3x3.transform = "b"

    def test_higher_order(self):
        weights = {
            0: [1.0, 1.0, 1.0],
            1: [1.0, 1.0, 1.0],
            2: [1.0, 1.0, 1.0],
            3: [1.0, 1.0, 1.0],
            4: [1.0, 1.0, 1.0, 1.0],
            5: [1.0, 1.0, 1.0],
            6: [1.0, 1.0, 1.0],
            7: [1.0, 1.0, 1.0],
            8: [1.0, 1.0, 1.0],
        }
        neighbors = {
            0: [4, 6, 2],
            1: [3, 5, 7],
            2: [8, 0, 4],
            3: [7, 1, 5],
            4: [8, 0, 2, 6],
            5: [1, 3, 7],
            6: [4, 0, 8],
            7: [3, 1, 5],
            8: [6, 2, 4],
        }
        wneighbs = {
            k: {neighb: weights[k][i] for i, neighb in enumerate(v)}
            for k, v in list(neighbors.items())
        }
        w2 = util.higher_order(self.w3x3, 2)
        test_wneighbs = {
            k: {ne: w2.weights[k][i] for i, ne in enumerate(v)}
            for k, v in list(w2.neighbors.items())
        }
        self.assertEqual(test_wneighbs, wneighbs)

    def test_histogram(self):
        hist = [
            (0, 1),
            (1, 1),
            (2, 4),
            (3, 20),
            (4, 57),
            (5, 44),
            (6, 36),
            (7, 15),
            (8, 7),
            (9, 1),
            (10, 6),
            (11, 0),
            (12, 2),
            (13, 0),
            (14, 0),
            (15, 1),
        ]
        self.assertEqual(self.w.histogram, hist)

    def test_id2i(self):
        id2i = {0: 0, 1: 1, 2: 2, 3: 3, 4: 4, 5: 5, 6: 6, 7: 7, 8: 8}
        self.assertEqual(self.w3x3.id2i, id2i)

    def test_id_order_set(self):
        w = W(neighbors={"a": ["b"], "b": ["a", "c"], "c": ["b"]})
        self.assertFalse(w.id_order_set)

    def test_islands(self):
        w = W(
            neighbors={"a": ["b"], "b": ["a", "c"], "c": ["b"], "d": []},
            silence_warnings=True,
        )
        self.assertEqual(w.islands, ["d"])
        self.assertEqual(self.w3x3.islands, [])

    def test_max_neighbors(self):
        w = W(
            neighbors={"a": ["b"], "b": ["a", "c"], "c": ["b"], "d": []},
            silence_warnings=True,
        )
        self.assertEqual(w.max_neighbors, 2)
        self.assertEqual(self.w3x3.max_neighbors, 4)

    def test_mean_neighbors(self):
        w = util.lat2W()
        self.assertEqual(w.mean_neighbors, 3.2)

    def test_min_neighbors(self):
        w = util.lat2W()
        self.assertEqual(w.min_neighbors, 2)

    def test_n(self):
        w = util.lat2W()
        self.assertEqual(w.n, 25)

    def test_nonzero(self):
        self.assertEqual(self.w3x3.nonzero, 24)

    def test_order(self):
        w = util.lat2W(3, 3)
        o = {
            0: [-1, 1, 2, 1, 2, 3, 2, 3, 0],
            1: [1, -1, 1, 2, 1, 2, 3, 2, 3],
            2: [2, 1, -1, 3, 2, 1, 0, 3, 2],
            3: [1, 2, 3, -1, 1, 2, 1, 2, 3],
            4: [2, 1, 2, 1, -1, 1, 2, 1, 2],
            5: [3, 2, 1, 2, 1, -1, 3, 2, 1],
            6: [2, 3, 0, 1, 2, 3, -1, 1, 2],
            7: [3, 2, 3, 2, 1, 2, 1, -1, 1],
            8: [0, 3, 2, 3, 2, 1, 2, 1, -1],
        }
        self.assertEqual(util.order(w), o)

    def test_pct_nonzero(self):
        self.assertEqual(self.w3x3.pct_nonzero, 29.62962962962963)

    def test_s0(self):
        self.assertEqual(self.w3x3.s0, 24.0)

    def test_s1(self):
        self.assertEqual(self.w3x3.s1, 48.0)

    def test_s2(self):
        self.assertEqual(self.w3x3.s2, 272.0)

    def test_s2array(self):
        s2a = np.array(
            [[16.0], [36.0], [16.0], [36.0], [64.0], [36.0], [16.0], [36.0], [16.0]]
        )
        NPTA3E(self.w3x3.s2array, s2a)

    def test_sd(self):
        self.assertEqual(self.w3x3.sd, 0.66666666666666663)

    def test_set_transform(self):
        w = util.lat2W(2, 2)
        self.assertEqual(w.transform, "O")
        self.assertEqual(w.weights[0], [1.0, 1.0])
        w.transform = "r"
        self.assertEqual(w.weights[0], [0.5, 0.5])

    def test_shimbel(self):
        d = {
            0: [-1, 1, 2, 1, 2, 3, 2, 3, 4],
            1: [1, -1, 1, 2, 1, 2, 3, 2, 3],
            2: [2, 1, -1, 3, 2, 1, 4, 3, 2],
            3: [1, 2, 3, -1, 1, 2, 1, 2, 3],
            4: [2, 1, 2, 1, -1, 1, 2, 1, 2],
            5: [3, 2, 1, 2, 1, -1, 3, 2, 1],
            6: [2, 3, 4, 1, 2, 3, -1, 1, 2],
            7: [3, 2, 3, 2, 1, 2, 1, -1, 1],
            8: [4, 3, 2, 3, 2, 1, 2, 1, -1],
        }
        self.assertEqual(util.shimbel(self.w3x3), d)

    def test_sparse(self):
        self.assertEqual(self.w3x3.sparse.nnz, 24)

    def test_trcW2(self):
        self.assertEqual(self.w3x3.trcW2, 24.0)

    def test_trcWtW(self):
        self.assertEqual(self.w3x3.trcWtW, 24.0)

    def test_trcWtW_WW(self):
        self.assertEqual(self.w3x3.trcWtW_WW, 48.0)


class TestWSP(unittest.TestCase):
    def setUp(self):
        self.w = psopen(examples.get_path("sids2.gal")).read()
        self.wsp = WSP(self.w.sparse, self.w.id_order)
        w3x3 = util.lat2W(3, 3)
        self.w3x3 = WSP(w3x3.sparse)

    def test_WSP(self):
        self.assertEqual(self.w.id_order, self.wsp.id_order)
        self.assertEqual(self.w.n, self.wsp.n)
        np.testing.assert_array_equal(
            self.w.sparse.todense(), self.wsp.sparse.todense()
        )

    def test_diagWtW_WW(self):
        NPTA3E(
            self.w3x3.diagWtW_WW,
            np.array([4.0, 6.0, 4.0, 6.0, 8.0, 6.0, 4.0, 6.0, 4.0]),
        )

    def test_trcWtW_WW(self):
        self.assertEqual(self.w3x3.trcWtW_WW, 48.0)

    def test_s0(self):
        self.assertEqual(self.w3x3.s0, 24.0)


if __name__ == "__main__":
    unittest.main()



"""
libpysal: Python Spatial Analysis Library (core)
================================================


Documentation
-------------
PySAL documentation is available in two forms: python docstrings and an html \
        webpage at http://pysal.org/

Available sub-packages
----------------------

cg
    Basic data structures and tools for Computational Geometry
examples
    Example data sets for testing and documentation
io
    Basic functions used by several sub-packages
weights
    Tools for creating and manipulating weights
"""
from . import cg
from . import io
from . import weights
from . import examples

from . import _version

__version__ = _version.get_versions()["version"]



import numpy as np
from ..fileio import FileIO as psopen
from ...common import pandas

# import pysal_examples
from ... import examples as pysal_examples
import unittest as ut

PANDAS_EXTINCT = pandas is None


class Test_Table(ut.TestCase):
    def setUp(self):
        self.filehandler = psopen(pysal_examples.get_path("columbus.dbf"))
        self.df = self.filehandler.to_df()
        self.filehandler.seek(0)
        self.shapefile = psopen(pysal_examples.get_path("columbus.shp"))
        self.csvhandler = psopen(pysal_examples.get_path("usjoin.csv"))
        self.csv_df = self.csvhandler.to_df()
        self.csvhandler.seek(0)

    @ut.skipIf(PANDAS_EXTINCT, "Missing pandas.")
    def test_to_df(self):
        for column in self.csv_df.columns:
            if column.lower() == "name":
                continue
            np.testing.assert_allclose(
                self.csvhandler.by_col(column), self.csv_df[column].values
            )
        for column in self.df.columns:
            if column == "geometry":
                continue
            np.testing.assert_allclose(self.filehandler.by_col(column), self.df[column])



import unittest
from ..wkt import WKTParser
from ....cg.shapes import Point, Chain, Polygon


class test_WKTParser(unittest.TestCase):
    def setUp(self):
        # Create some Well-Known Text objects
        self.wktPOINT = "POINT(6 10)"
        self.wktLINESTRING = "LINESTRING(3 4,10 50,20 25)"
        self.wktPOLYGON = "POLYGON((1 1,5 1,5 5,1 5,1 1),(2 2, 3 2, 3 3, 2 3,2 2))"
        self.unsupported = [
            "MULTIPOINT(3.5 5.6,4.8 10.5)",
            "MULTILINESTRING((3 4,10 50,20 25),(-5 -8,-10 -8,-15 -4))",
            "MULTIPOLYGON(((1 1,5 1,5 5,1 5,1 1),(2 2, 3 2, 3 3, 2 3,2 2)),((3 3,6 2,6 4,3 3)))",
            "GEOMETRYCOLLECTION(POINT(4 6),LINESTRING(4 6,7 10))",
            "POINT ZM (1 1 5 60)",
            "POINT M (1 1 80)",
        ]
        self.empty = ["POINT EMPTY", "MULTIPOLYGON EMPTY"]
        self.parser = WKTParser()

    def test_Point(self):
        pt = self.parser(self.wktPOINT)
        self.assertTrue(issubclass(type(pt), Point))
        self.assertEqual(pt[:], (6.0, 10.0))

    def test_LineString(self):
        line = self.parser(self.wktLINESTRING)
        self.assertTrue(issubclass(type(line), Chain))
        parts = [[pt[:] for pt in part] for part in line.parts]
        self.assertEqual(parts, [[(3.0, 4.0), (10.0, 50.0), (20.0, 25.0)]])
        self.assertEqual(line.len, 73.455384532199886)

    def test_Polygon(self):
        poly = self.parser(self.wktPOLYGON)
        self.assertTrue(issubclass(type(poly), Polygon))
        parts = [[pt[:] for pt in part] for part in poly.parts]
        self.assertEqual(
            parts,
            [
                [(1.0, 1.0), (1.0, 5.0), (5.0, 5.0), (5.0, 1.0), (1.0, 1.0)],
                [(2.0, 2.0), (2.0, 3.0), (3.0, 3.0), (3.0, 2.0), (2.0, 2.0)],
            ],
        )
        self.assertEqual(poly.centroid, (2.9705882352941178, 2.9705882352941178))
        self.assertEqual(poly.area, 17.0)

    def test_fromWKT(self):
        for wkt in self.unsupported:
            self.assertRaises(NotImplementedError, self.parser.fromWKT, wkt)
        for wkt in self.empty:
            self.assertEqual(self.parser.fromWKT(wkt), None)
        self.assertEqual(self.parser.__call__, self.parser.fromWKT)


if __name__ == "__main__":
    unittest.main()



"""
PySAL ShapeFile Reader and Writer based on pure python shapefile module.

"""

__author__ = "Charles R Schmidt <schmidtc@gmail.com>"
__credits__ = "Copyright (c) 2009 Charles R. Schmidt"
__all__ = ["PurePyShpWrapper"]

from .. import fileio
from ..util import shp_file
from ... import cg
from warnings import warn

STRING_TO_TYPE = {
    "POLYGON": cg.Polygon,
    "POINT": cg.Point,
    "POINTM": cg.Point,
    "POINTZ": cg.Point,
    "ARC": cg.Chain,
    "POLYGONZ": cg.Polygon,
}

# build the reverse map
# for key,value in STRING_TO_TYPE.items():
#    TYPE_TO_STRING[value] = key

TYPE_TO_STRING = {
    cg.Polygon: "POLYGON",
    cg.Point: "POINT",
    cg.Chain: "ARC",
}


class PurePyShpWrapper(fileio.FileIO):
    """FileIO handler for ESRI ShapeFiles.

    Notes
    -----
    
    This class wraps ``_pyShpIO``'s ``shp_file`` class with the PySAL `FileIO` API.
    shp_file can be used without PySAL.

    Examples
    --------

    >>> import tempfile
    >>> f = tempfile.NamedTemporaryFile(suffix='.shp')
    >>> fname = f.name
    >>> f.close()
    
    >>> import libpysal
    >>> i = libpysal.io.open(libpysal.examples.get_path('10740.shp'),'r')
    >>> o = libpysal.io.open(fname,'w')
    
    >>> for shp in i:
    ...     o.write(shp)
    >>> o.close()
    
    >>> one = libpysal.io.open(libpysal.examples.get_path('10740.shp'),'rb').read()
    >>> two = libpysal.io.open(fname,'rb').read()
    >>> one[0].centroid == two[0].centroid
    True
    
    >>> one = libpysal.io.open(libpysal.examples.get_path('10740.shx'),'rb').read()
    >>> two = libpysal.io.open(fname[:-1]+'x','rb').read()
    >>> one[0].centroid == two[0].centroid
    True
    
    >>> import os
    >>> os.remove(fname); os.remove(fname.replace('.shp','.shx'))

    """

    FORMATS = ["shp", "shx"]
    MODES = ["w", "r", "wb", "rb"]

    def __init__(self, *args, **kwargs):
        fileio.FileIO.__init__(self, *args, **kwargs)
        self.dataObj = None
        if self.mode == "r" or self.mode == "rb":
            self.__open()
        elif self.mode == "w" or self.mode == "wb":
            self.__create()

    def __len__(self) -> int:
        if self.dataObj != None:
            return len(self.dataObj)
        else:
            return 0

    def __open(self):
        """

        Raises
        ------
        TypeError
            Raised when an invalid shape is passed in.

        """

        self.dataObj = shp_file(self.dataPath)
        self.header = self.dataObj.header
        self.bbox = self.dataObj.bbox

        try:
            self.type = STRING_TO_TYPE[self.dataObj.type()]
        except KeyError:
            msg = "%s does not support shapes of type: %s."
            msg = msg % (self.__class__.__name__, self.dataObj.type())
            raise TypeError(msg)

    def __create(self):
        self.write = self.__firstWrite

    def __firstWrite(self, shape):
        """

        Parameters
        ----------
        shape : libpysal.cg.{Point, Chain, Polygon}
            Geometric shape.

        """

        self.type = TYPE_TO_STRING[type(shape)]
        if self.type == "POINT":
            if len(shape) == 3:
                self.type = "POINTM"
            if len(shape) == 4:
                self.type = "POINTZ"
        self.dataObj = shp_file(self.dataPath, "w", self.type)
        self.write = self.__writer
        self.write(shape)

    def __writer(self, shape):
        """

        Parameters
        ----------
        shape : libpysal.cg.{Point, Chain, Polygon}
            Geometric shape.

        Raises
        ------
        TypeError
            Raised when an invalid shape is passed in.

        """

        if TYPE_TO_STRING[type(shape)] != self.type:
            raise TypeError("This file only supports %s type shapes." % self.type)

        rec = {}
        rec["Shape Type"] = shp_file.SHAPE_TYPES[self.type]

        if self.type == "POINT":
            rec["X"] = shape[0]
            rec["Y"] = shape[1]
            if len(shape) > 2:
                rec["M"] = shape[2]
            if len(shape) > 3:
                rec["Z"] = shape[3]
            shape = rec
        else:
            rec["BBOX Xmin"] = shape.bounding_box.left
            rec["BBOX Ymin"] = shape.bounding_box.lower
            rec["BBOX Xmax"] = shape.bounding_box.right
            rec["BBOX Ymax"] = shape.bounding_box.upper
            if self.type == "POLYGON":
                holes = [hole[::-1] for hole in shape.holes if hole]
                # holes should be in CCW order
                rec["NumParts"] = len(shape.parts) + len(holes)
                all_parts = shape.parts + holes
            else:
                rec["NumParts"] = len(shape.parts)
                all_parts = shape.parts
            partsIndex = [0]
            for l in [len(part) for part in all_parts][:-1]:
                partsIndex.append(partsIndex[-1] + l)
            rec["Parts Index"] = partsIndex
            verts = sum(all_parts, [])
            verts = [(x, y) for x, y in verts]
            rec["NumPoints"] = len(verts)
            rec["Vertices"] = verts
        self.dataObj.add_shape(rec)
        self.pos += 1

    def _read(self):
        """

        Returns
        -------
        shape : libpysal.cg.{Point, Chain, Polygon}
            Geometric shape.

        """
        try:
            rec = self.dataObj.get_shape(self.pos)
        except IndexError:
            return None

        self.pos += 1

        if self.dataObj.type() == "POINT":
            shp = self.type((rec["X"], rec["Y"]))
        elif self.dataObj.type() == "POINTZ":
            shp = self.type((rec["X"], rec["Y"]))
            shp.Z = rec["Z"]
            shp.M = rec["M"]
        else:
            if rec["NumParts"] > 1:
                partsIndex = list(rec["Parts Index"])
                partsIndex.append(None)
                parts = [
                    rec["Vertices"][partsIndex[i] : partsIndex[i + 1]]
                    for i in range(rec["NumParts"])
                ]
                if self.dataObj.type() == "POLYGON":
                    is_cw = [cg.is_clockwise(part) for part in parts]
                    vertices = [part for part, cw in zip(parts, is_cw) if cw]
                    holes = [part for part, cw in zip(parts, is_cw) if not cw]
                    if not holes:
                        holes = None
                    shp = self.type(vertices, holes)
                else:
                    vertices = parts
                    shp = self.type(vertices)
            elif rec["NumParts"] == 1:
                vertices = rec["Vertices"]
                if self.dataObj.type() == "POLYGON" and not cg.is_clockwise(vertices):

                    ### SHAPEFILE WARNING: Polygon %d topology has been fixed. (ccw -> cw)
                    msg = "SHAPEFILE WARNING: Polygon %d "
                    msg += "topology has been fixed. (ccw -> cw)."
                    msg = msg % self.pos
                    warn(msg, RuntimeWarning)
                    print(msg)

                shp = self.type(vertices)
            else:
                warn("Polygon %d has zero parts." % self.pos, RuntimeWarning)
                shp = self.type([[]])
                # raise ValueError, "Polygon %d has zero parts"%self.pos

        if self.ids:
            # shp IDs start at 1.
            shp.id = self.rIds[self.pos - 1]
        else:
            # shp IDs start at 1.
            shp.id = self.pos

        return shp

    def close(self):
        self.dataObj.close()
        fileio.FileIO.close(self)



import os
import unittest
import numpy as np
from ..contiguity import Rook, Voronoi
from .. import user
from ... import examples


class Testuser(unittest.TestCase):
    def test_min_threshold_dist_from_shapefile(self):
        f = examples.get_path("columbus.shp")
        min_d = user.min_threshold_dist_from_shapefile(f)
        self.assertAlmostEqual(min_d, 0.61886415807685413)

    def test_build_lattice_shapefile(self):
        of = "lattice.shp"
        user.build_lattice_shapefile(20, 20, of)
        w = Rook.from_shapefile(of)
        self.assertEqual(w.n, 400)
        os.remove("lattice.shp")
        os.remove("lattice.shx")


suite = unittest.TestLoader().loadTestsFromTestCase(Testuser)

if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite)



import unittest as ut
from .. import tabular as ta
from ....common import RTOL, ATOL, pandas, requires as _requires
from ....examples import get_path
from ...shapes import Polygon
from ....io import geotable as pdio
from ... import ops as GIS
import numpy as np

try:
    import shapely as shp
except ImportError:
    shp = None

PANDAS_EXTINCT = pandas is None
SHAPELY_EXTINCT = shp is None


@ut.skipIf(PANDAS_EXTINCT or SHAPELY_EXTINCT, "Missing pandas or shapely.")
class Test_Tabular(ut.TestCase):
    def setUp(self):
        import pandas as pd

        self.columbus = pdio.read_files(get_path("columbus.shp"))
        grid = [
            Polygon([(0, 0), (0, 1), (1, 1), (1, 0)]),
            Polygon([(0, 1), (0, 2), (1, 2), (1, 1)]),
            Polygon([(1, 2), (2, 2), (2, 1), (1, 1)]),
            Polygon([(1, 1), (2, 1), (2, 0), (1, 0)]),
        ]
        regime = [0, 0, 1, 1]
        ids = list(range(4))
        data = np.array((regime, ids)).T
        self.exdf = pd.DataFrame(data, columns=["regime", "ids"])
        self.exdf["geometry"] = grid

    @_requires("geopandas")
    def test_round_trip(self):
        import geopandas as gpd
        import pandas as pd

        geodf = GIS.tabular.to_gdf(self.columbus)
        self.assertIsInstance(geodf, gpd.GeoDataFrame)
        new_df = GIS.tabular.to_df(geodf)
        self.assertIsInstance(new_df, pd.DataFrame)
        for new, old in zip(new_df.geometry, self.columbus.geometry):
            self.assertEqual(new, old)

    def test_spatial_join(self):
        pass

    def test_spatial_overlay(self):
        pass

    def test_dissolve(self):
        out = GIS.tabular.dissolve(self.exdf, by="regime")
        self.assertEqual(out[0].area, 2.0)
        self.assertEqual(out[1].area, 2.0)

        answer_vertices0 = set([(0, 0), (0, 1), (0, 2), (1, 2), (1, 1), (1, 0), (0, 0)])
        answer_vertices1 = set([(2, 1), (2, 0), (1, 0), (1, 1), (1, 2), (2, 2), (2, 1)])

        s0 = set([tuple(map(int,t)) for t in out[0].vertices])
        s1 = set([tuple(map(int,t)) for t in out[1].vertices])

        self.assertTrue(s0==answer_vertices0)
        self.assertTrue(s1==answer_vertices1)


    def test_clip(self):
        pass

    def test_erase(self):
        pass

    def test_union(self):
        new_geom = GIS.tabular.union(self.exdf)
        self.assertEqual(new_geom.area, 4)

    def test_intersection(self):
        pass

    def test_symmetric_difference(self):
        pass

    def test_difference(self):
        pass



"""
Spatial weights for spatial interaction including contiguity OD weights (ODW),
network based weights (netW), and distance-decay based vector weights (vecW).

"""

__author__ = "Taylor Oshan  <tayoshan@gmail.com> "

from scipy.sparse import kron
from .weights import W, WSP
from .distance import DistanceBand
from collections import OrderedDict


def ODW(Wo, Wd, transform="r", silence_warnings=True):
    """
    Constructs an o*d by o*d origin-destination style spatial weight for o*d
    flows using standard spatial weights on o origins and d destinations. Input
    spatial weights must be binary or able to be sutiably transformed to binary.

    Parameters
    ----------
    Wo          : W object for origin locations
                  o x o spatial weight object amongst o origins

    Wd          : W object for destination locations
                  d x d spatial weight object amongst d destinations

    transform   : Transformation for standardization of final OD spatial weight; default
                  is 'r' for row standardized

    Returns
    -------
    W           : spatial contiguity W object for assocations between flows
                 o*d x o*d spatial weight object amongst o*d flows between o
                 origins and d destinations

    Examples
    --------

    >>> import libpysal
    >>> O = libpysal.weights.lat2W(2,2)
    >>> D = libpysal.weights.lat2W(2,2)
    >>> OD = libpysal.weights.ODW(O,D)
    >>> OD.weights[0]
    [0.25, 0.25, 0.25, 0.25]
    >>> OD.neighbors[0]
    [5, 6, 9, 10]
    >>> OD.full()[0][0]
    array([0.  , 0.  , 0.  , 0.  , 0.  , 0.25, 0.25, 0.  , 0.  , 0.25, 0.25,
           0.  , 0.  , 0.  , 0.  , 0.  ])

    """
    if Wo.transform != "b":
        try:
            Wo.tranform = "b"
        except:
            raise AttributeError(
                "Wo is not binary and cannot be transformed to "
                "binary. Wo must be binary or suitably transformed to binary."
            )
    if Wd.transform != "b":
        try:
            Wd.tranform = "b"
        except:
            raise AttributeError(
                "Wd is not binary and cannot be transformed to "
                "binary. Wd must be binary or suitably transformed to binary."
            )
    Wo = Wo.sparse
    Wo.eliminate_zeros()
    Wd = Wd.sparse
    Wd.eliminate_zeros()
    Ww = kron(Wo, Wd, format="csr")
    Ww.eliminate_zeros()
    Ww = WSP(Ww).to_W(silence_warnings=silence_warnings)
    Ww.transform = transform
    return Ww


def netW(link_list, share="A", transform="r", **kwargs):
    """
    Create a network-contiguity based weight object based on different nodal
    relationships encoded in a network.

    Parameters
    ----------
    link_list   : list
                  of tuples where each tuple is of the form (o,d) where o is an
                  origin id and d is a destination id

    share       : string
                  denoting how to define the nodal relationship used to determine neighboring edges; defualt is 'A' for any shared nodes between two network edges; options include: O a shared origin node; D a shared destination node; OD; a shared origin or a shared destination node; C a shared node that is the destination of the first edge and the origin of the second edge - i.e., a directed chain is formed moving from edge one to edge two.

    transform   : Transformation for standardization of final OD spatial weight; default
                  is 'r' for row standardized
    **kwargs    : keyword arguments
                  optional arguments for :class:`pysal.weights.W`


    Returns
    -------
     W          : nodal contiguity W object for networkd edges or flows
                  W Object representing the binary adjacency of the network edges
                  given a definition of nodal relationshilibpysal.weights.spintW.

    Examples
    --------
    >>> import libpysal
    >>> links = [('a','b'), ('a','c'), ('a','d'), ('c','d'), ('c', 'b'), ('c','a')]
    >>> O = libpysal.weights.netW(links, share='O')
    >>> O.neighbors[('a', 'b')]
    [('a', 'c'), ('a', 'd')]
    >>> OD = libpysal.weights.netW(links, share='OD')
    >>> OD.neighbors[('a', 'b')]
    [('a', 'c'), ('a', 'd'), ('c', 'b')]
    >>> any_common = libpysal.weights.netW(links, share='A')
    >>> any_common.neighbors[('a', 'b')]
    [('a', 'c'), ('a', 'd'), ('c', 'b'), ('c', 'a')]

    """
    neighbors = {}
    neighbors = OrderedDict()
    edges = link_list
    for key in edges:
        neighbors[key] = []
        for neigh in edges:
            if key == neigh:
                continue
            if share.upper() == "OD":
                if key[0] == neigh[0] or key[1] == neigh[1]:
                    neighbors[key].append(neigh)
            elif share.upper() == "O":
                if key[0] == neigh[0]:
                    neighbors[key].append(neigh)
            elif share.upper() == "D":
                if key[1] == neigh[1]:
                    neighbors[key].append(neigh)
            elif share.upper() == "C":
                if key[1] == neigh[0]:
                    neighbors[key].append(neigh)
            elif share.upper() == "A":
                if (
                    key[0] == neigh[0]
                    or key[0] == neigh[1]
                    or key[1] == neigh[0]
                    or key[1] == neigh[1]
                ):
                    neighbors[key].append(neigh)
            else:
                raise AttributeError(
                    "Parameter 'share' must be 'O', 'D'," " 'OD', or 'C'"
                )
    netW = W(neighbors, **kwargs)
    netW.tranform = transform
    return netW


def vecW(
    origin_x,
    origin_y,
    dest_x,
    dest_y,
    threshold,
    p=2,
    alpha=-1.0,
    binary=True,
    ids=None,
    build_sp=False,
    **kwargs
):
    """
    Distance-based spatial weight for vectors that is computed using a
    4-dimensional distance between the origin x,y-coordinates and the
    destination x,y-coordinates

    Parameters
    ----------
    origin_x   : list or array
                 of vector origin x-coordinates
    origin_y   : list or array
                 of vector origin y-coordinates
    dest_x     : list or array
                 of vector destination x-coordinates
    dest_y     : list or array
                 of vector destination y-coordinates
    threshold  : float
                 distance band
    p          : float
                 Minkowski p-norm distance metric parameter:
                 1<=p<=infinity
                 2: Euclidean distance
                 1: Manhattan distance
    binary     : boolean
                 If true w_{ij}=1 if d_{i,j}<=threshold, otherwise w_{i,j}=0
                 If false wij=dij^{alpha}
    alpha      : float
                 distance decay parameter for weight (default -1.0)
                 if alpha is positive the weights will not decline with
                 distance. If binary is True, alpha is ignored

    ids         : list
                  values to use for keys of the neighbors and weights dicts
    build_sp    : boolean
                  True to build sparse distance matrix and false to build dense
                  distance matrix; significant speed gains may be obtained
                  dending on the sparsity of the of distance_matrix and
                  threshold that is applied
    **kwargs    : keyword arguments
                  optional arguments for :class:`pysal.weights.W`


    Returns
    ------
    W           : DistanceBand W object that uses 4-dimenional distances between
                  vectors origin and destination coordinates.

    Examples
    --------
    >>> import libpysal
    >>> x1 = [5,6,3]
    >>> y1 = [1,8,5]
    >>> x2 = [2,4,9]
    >>> y2 = [3,6,1]
    >>> W1 = libpysal.weights.vecW(x1, y1, x2, y2, threshold=999)
    >>> list(W1.neighbors[0])
    [1, 2]
    >>> W2 = libpysal.weights.vecW(x1, y2, x1, y2, threshold=8.5)
    >>> list(W2.neighbors[0])
    [1, 2]

    """
    data = list(zip(origin_x, origin_y, dest_x, dest_y))
    W = DistanceBand(
        data,
        threshold=threshold,
        p=p,
        binary=binary,
        alpha=alpha,
        ids=ids,
        build_sp=False,
        **kwargs
    )
    return W


def mat2L(edge_matrix):
    """
    Convert a matrix denoting network connectivity (edges or flows) to a list
    denoting edges

    Parameters
    ----------
    edge_matrix   : array
                    where rows denote network edge origins, columns denote
                    network edge destinations, and non-zero entries denote the
                    existence of an edge between a given origin and destination

    Returns
    -------
     edge_list    : list
                    of tuples where each tuple is of the form (o,d) where o is an
                    origin id and d is a destination id

    """
    if len(edge_matrix.shape) != 2:
        raise AttributeError(
            "Matrix of network edges should be two dimensions"
            "with edge origins on one axis and edge destinations on the"
            "second axis with non-zero matrix entires denoting an edge"
            "between and origin and destination"
        )
    edge_list = []
    rows, cols = edge_matrix.shape
    for row in range(rows):
        for col in range(cols):
            if edge_matrix[row, col] != 0:
                edge_list.append((row, col))
    return edge_list



import warnings

warnings.filterwarnings(action="ignore", message=".*__builtin__.file size changed.*")
from . import gwt
from . import gal
from . import dat
from . import pyShpIO
from . import wkt
from . import geoda_txt
from . import csvWrapper
from . import pyDbfIO
from . import arcgis_dbf
from . import arcgis_swm
from . import arcgis_txt
from . import dat
from . import geobugs_txt
from . import mat
from . import mtx
from . import stata_txt
from . import wk1

try:
    from . import db
except:
    warnings.warn("SQLAlchemy and Geomet not installed, database I/O disabled")



import itertools

import numpy

from ..cg import voronoi_frames
from ..io.fileio import FileIO
from ._contW_lists import ContiguityWeightsLists
from .util import get_ids, get_points_array
from .weights import WSP, W
from .raster import da2W, da2WSP

try:
    from shapely.geometry import Point as shapely_point
    from ..cg.shapes import Point as pysal_point

    point_type = (shapely_point, pysal_point)
except ImportError:
    from ..cg.shapes import Point as point_type

WT_TYPE = {"rook": 2, "queen": 1}  # for _contW_Binning

__author__ = "Sergio J. Rey <srey@asu.edu> , Levi John Wolf <levi.john.wolf@gmail.com>"

__all__ = ["Rook", "Queen", "Voronoi"]


class Rook(W):
    """
    Construct a weights object from a collection of pysal polygons that share at least one edge.

    Parameters
    ----------
    polygons    : list
                a collection of PySAL shapes to build weights from
    ids         : list
                a list of names to use to build the weights
    **kw        : keyword arguments
                optional arguments for :class:`pysal.weights.W`

    See Also
    ---------
    :class:`libpysal.weights.weights.W`
    """

    def __init__(self, polygons, **kw):
        criterion = "rook"
        ids = kw.pop("ids", None)
        polygons, backup = itertools.tee(polygons)
        first_shape = next(iter(backup))
        if isinstance(first_shape, point_type):
            polygons, vertices = voronoi_frames(get_points_array(polygons))
            polygons = list(polygons.geometry)
        neighbors, ids = _build(polygons, criterion=criterion, ids=ids)
        W.__init__(self, neighbors, ids=ids, **kw)

    @classmethod
    def from_shapefile(cls, filepath, idVariable=None, full=False, **kwargs):
        """
        Rook contiguity weights from a polygon shapefile.

        Parameters
        ----------

        shapefile : string
                    name of polygon shapefile including suffix.
        sparse    : boolean
                    If True return WSP instance
                    If False return W instance

        Returns
        -------

        w          : W
                     instance of spatial weights

        Examples
        --------
        >>> from libpysal.weights import Rook
        >>> import libpysal
        >>> wr=Rook.from_shapefile(libpysal.examples.get_path("columbus.shp"), "POLYID")
        >>> "%.3f"%wr.pct_nonzero
        '8.330'
        >>> wr=Rook.from_shapefile(libpysal.examples.get_path("columbus.shp"), sparse=True)
        >>> pct_sp = wr.sparse.nnz *1. / wr.n**2
        >>> "%.3f"%pct_sp
        '0.083'

        Notes
        -----

        Rook contiguity defines as neighbors any pair of polygons that share a
        common edge in their polygon definitions.

        See Also
        --------
        :class:`libpysal.weights.weights.W`
        :class:`libpysal.weights.contiguity.Rook`
        """
        sparse = kwargs.pop("sparse", False)
        if idVariable is not None:
            ids = get_ids(filepath, idVariable)
        else:
            ids = None
        w = cls(FileIO(filepath), ids=ids, **kwargs)
        w.set_shapefile(filepath, idVariable=idVariable, full=full)
        if sparse:
            w = w.to_WSP()
        return w

    @classmethod
    def from_iterable(cls, iterable, sparse=False, **kwargs):
        """
        Construct a weights object from a collection of arbitrary polygons. This
        will cast the polygons to PySAL polygons, then build the W.

        Parameters
        ----------
        iterable    : iterable
                      a collection of of shapes to be cast to PySAL shapes. Must
                      support iteration. Can be either Shapely or PySAL shapes.
        **kw        : keyword arguments
                      optional arguments for  :class:`pysal.weights.W`
        See Also
        --------
        :class:`libpysal.weights.weights.W`
        :class:`libpysal.weights.contiguity.Rook`
        """
        new_iterable = iter(iterable)
        w = cls(new_iterable, **kwargs)
        if sparse:
            w = WSP.from_W(w)
        return w

    @classmethod
    def from_dataframe(
        cls, df, geom_col=None, idVariable=None, ids=None, id_order=None, **kwargs
    ):
        """
        Construct a weights object from a pandas dataframe with a geometry
        column. This will cast the polygons to PySAL polygons, then build the W
        using ids from the dataframe.

        Parameters
        ----------
        df          : DataFrame
                      a :class: `pandas.DataFrame` containing geometries to use
                      for spatial weights
        geom_col    : string
                      the name of the column in `df` that contains the
                      geometries. Defaults to active geometry column.
        idVariable  : string
                      the name of the column to use as IDs. If nothing is
                      provided, the dataframe index is used
        ids         : list
                      a list of ids to use to index the spatial weights object.
                      Order is not respected from this list.
        id_order    : list
                      an ordered list of ids to use to index the spatial weights
                      object. If used, the resulting weights object will iterate
                      over results in the order of the names provided in this
                      argument. 

        See Also
        --------
        :class:`libpysal.weights.weights.W`
        :class:`libpysal.weights.contiguity.Rook`
        """
        if geom_col is None:
            geom_col = df.geometry.name
        if id_order is not None:
            if id_order is True and ((idVariable is not None) or (ids is not None)):
                # if idVariable is None, we want ids. Otherwise, we want the
                # idVariable column
                id_order = list(df.get(idVariable, ids))
            else:
                id_order = df.get(id_order, ids)
        elif idVariable is not None:
            ids = df.get(idVariable).tolist()
        elif isinstance(ids, str):
            ids = df.get(ids).tolist()
        return cls.from_iterable(
            df[geom_col].tolist(), ids=ids, id_order=id_order, **kwargs
        )

    @classmethod
    def from_xarray(
        cls,
        da,
        z_value=None,
        coords_labels={},
        k=1,
        include_nodata=False,
        n_jobs=1,
        sparse=True,
        **kwargs,
    ):
        """
        Construct a weights object from a xarray.DataArray with an additional
        attribute index containing coordinate values of the raster
        in the form of Pandas.Index/MultiIndex.

        Parameters
        ----------
        da : xarray.DataArray
            Input 2D or 3D DataArray with shape=(z, y, x)
        z_value : int/string/float
            Select the z_value of 3D DataArray with multiple layers.
        coords_labels : dictionary
            Pass dimension labels for coordinates and layers if they do not
            belong to default dimensions, which are (band/time, y/lat, x/lon)
            e.g. coords_labels = {"y_label": "latitude", "x_label": "longitude", "z_label": "year"}
            Default is {} empty dictionary.
        sparse : boolean
            type of weight object. Default is True. For libpysal.weights.W, sparse = False
        k : int
            Order of contiguity, this will select all neighbors upto kth order.
            Default is 1.
        include_nodata : boolean
            If True, missing values will be assumed as non-missing when
            selecting higher_order neighbors, Default is False
        n_jobs : int
            Number of cores to be used in the sparse weight construction. If -1,
            all available cores are used. Default is 1.
        **kwargs : keyword arguments
            optional arguments passed when sparse = False

        Returns
        -------
        w : libpysal.weights.W/libpysal.weights.WSP
            instance of spatial weights class W or WSP with an index attribute 

        Notes
        -----
        1. Lower order contiguities are also selected.
        2. Returned object contains `index` attribute that includes a
           `Pandas.MultiIndex` object from the DataArray.

        See Also
        --------
        :class:`libpysal.weights.weights.W`
        :class:`libpysal.weights.weights.WSP`
        """
        if sparse:
            w = da2WSP(da, "rook", z_value, coords_labels, k, include_nodata)
        else:
            w = da2W(da, "rook", z_value, coords_labels, k, include_nodata, **kwargs)
        return w


class Queen(W):
    """
    Construct a weights object from a collection of pysal polygons that share at least one vertex.

    Parameters
    ----------
    polygons    : list
                  a collection of PySAL shapes to build weights from
    ids         : list
                  a list of names to use to build the weights
    **kw        : keyword arguments
                  optional arguments for :class:`pysal.weights.W`

    See Also
    --------
    :class:`libpysal.weights.weights.W`
    """

    def __init__(self, polygons, **kw):
        criterion = "queen"
        ids = kw.pop("ids", None)
        polygons, backup = itertools.tee(polygons)
        first_shape = next(iter(backup))
        if isinstance(first_shape, point_type):
            polygons, vertices = voronoi_frames(get_points_array(polygons))
            polygons = list(polygons.geometry)
        neighbors, ids = _build(polygons, criterion=criterion, ids=ids)
        W.__init__(self, neighbors, ids=ids, **kw)

    @classmethod
    def from_shapefile(cls, filepath, idVariable=None, full=False, **kwargs):
        """
        Queen contiguity weights from a polygon shapefile.

        Parameters
        ----------

        shapefile   : string
                      name of polygon shapefile including suffix.
        idVariable  : string
                      name of a column in the shapefile's DBF to use for ids.
        sparse      : boolean
                      If True return WSP instance
                      If False return W instance
        Returns
        -------

        w            : W
                       instance of spatial weights

        Examples
        --------
        >>> from libpysal.weights import Queen
        >>> import libpysal
        >>> wq=Queen.from_shapefile(libpysal.examples.get_path("columbus.shp"))
        >>> "%.3f"%wq.pct_nonzero
        '9.829'
        >>> wq=Queen.from_shapefile(libpysal.examples.get_path("columbus.shp"),"POLYID")
        >>> "%.3f"%wq.pct_nonzero
        '9.829'
        >>> wq=Queen.from_shapefile(libpysal.examples.get_path("columbus.shp"), sparse=True)
        >>> pct_sp = wq.sparse.nnz *1. / wq.n**2
        >>> "%.3f"%pct_sp
        '0.098'

        Notes

        Queen contiguity defines as neighbors any pair of polygons that share at
        least one vertex in their polygon definitions.

        See Also
        --------
        :class:`libpysal.weights.weights.W`
        :class:`libpysal.weights.contiguity.Queen`
        """
        sparse = kwargs.pop("sparse", False)
        if idVariable is not None:
            ids = get_ids(filepath, idVariable)
        else:
            ids = None
        w = cls(FileIO(filepath), ids=ids, **kwargs)
        w.set_shapefile(filepath, idVariable=idVariable, full=full)
        if sparse:
            w = w.to_WSP()
        return w

    @classmethod
    def from_iterable(cls, iterable, sparse=False, **kwargs):
        """
        Construct a weights object from a collection of arbitrary polygons. This
        will cast the polygons to PySAL polygons, then build the W.

        Parameters
        ----------
        iterable    : iterable
                      a collection of of shapes to be cast to PySAL shapes. Must
                      support iteration. Contents may either be a shapely or PySAL shape.
        **kw        : keyword arguments
                      optional arguments for  :class:`pysal.weights.W`
        See Also
        ---------
        :class:`libpysal.weights.weights.W`
        :class:`libpysal.weights.contiguiyt.Queen`
        """
        new_iterable = iter(iterable)
        w = cls(new_iterable, **kwargs)
        if sparse:
            w = WSP.from_W(w)
        return w

    @classmethod
    def from_dataframe(cls, df, geom_col=None, **kwargs):
        """
        Construct a weights object from a pandas dataframe with a geometry
        column. This will cast the polygons to PySAL polygons, then build the W
        using ids from the dataframe.

        Parameters
        ----------
        df          : DataFrame
                      a :class: `pandas.DataFrame` containing geometries to use
                      for spatial weights
        geom_col    : string
                      the name of the column in `df` that contains the
                      geometries. Defaults to active geometry column
        idVariable  : string
                      the name of the column to use as IDs. If nothing is
                      provided, the dataframe index is used
        ids         : list
                      a list of ids to use to index the spatial weights object.
                      Order is not respected from this list.
        id_order    : list
                      an ordered list of ids to use to index the spatial weights
                      object. If used, the resulting weights object will iterate
                      over results in the order of the names provided in this
                      argument. 

        See Also
        --------
        :class:`libpysal.weights.weights.W`
        :class:`libpysal.weights.contiguity.Queen`
        """
        idVariable = kwargs.pop("idVariable", None)
        ids = kwargs.pop("ids", None)
        id_order = kwargs.pop("id_order", None)
        if geom_col is None:
            geom_col = df.geometry.name
        if id_order is not None:
            if id_order is True and ((idVariable is not None) or (ids is not None)):
                # if idVariable is None, we want ids. Otherwise, we want the
                # idVariable column
                ids = list(df.get(idVariable, ids))
                id_order = ids
            elif isinstance(id_order, str):
                ids = df.get(id_order, ids)
                id_order = ids
        elif idVariable is not None:
            ids = df.get(idVariable).tolist()
        elif isinstance(ids, str):
            ids = df.get(ids).tolist()
        w = cls.from_iterable(
            df[geom_col].tolist(), ids=ids, id_order=id_order, **kwargs
        )
        return w

    @classmethod
    def from_xarray(
        cls,
        da,
        z_value=None,
        coords_labels={},
        k=1,
        include_nodata=False,
        n_jobs=1,
        sparse=True,
        **kwargs,
    ):
        """
        Construct a weights object from a xarray.DataArray with an additional
        attribute index containing coordinate values of the raster
        in the form of Pandas.Index/MultiIndex.

        Parameters
        ----------
        da : xarray.DataArray
            Input 2D or 3D DataArray with shape=(z, y, x)
        z_value : int/string/float
            Select the z_value of 3D DataArray with multiple layers.
        coords_labels : dictionary
            Pass dimension labels for coordinates and layers if they do not
            belong to default dimensions, which are (band/time, y/lat, x/lon)
            e.g. coords_labels = {"y_label": "latitude", "x_label": "longitude", "z_label": "year"}
            Default is {} empty dictionary.
        sparse : boolean
            type of weight object. Default is True. For libpysal.weights.W, sparse = False
        k : int
            Order of contiguity, this will select all neighbors upto kth order.
            Default is 1.
        include_nodata : boolean
            If True, missing values will be assumed as non-missing when
            selecting higher_order neighbors, Default is False
        n_jobs : int
            Number of cores to be used in the sparse weight construction. If -1,
            all available cores are used. Default is 1.
        **kwargs : keyword arguments
            optional arguments passed when sparse = False

        Returns
        -------
        w : libpysal.weights.W/libpysal.weights.WSP
            instance of spatial weights class W or WSP with an index attribute 

        Notes
        -----
        1. Lower order contiguities are also selected.
        2. Returned object contains `index` attribute that includes a
           `Pandas.MultiIndex` object from the DataArray.

        See Also
        --------
        :class:`libpysal.weights.weights.W`
        :class:`libpysal.weights.weights.WSP`
        """
        if sparse:
            w = da2WSP(da, "queen", z_value, coords_labels, k, include_nodata)
        else:
            w = da2W(da, "queen", z_value, coords_labels, k, include_nodata, **kwargs)
        return w


def Voronoi(points, criterion="rook", clip="ahull", **kwargs):
    """
    Voronoi weights for a 2-d point set


    Points are Voronoi neighbors if their polygons share an edge or vertex.


    Parameters
    ----------

    points      : array
                  (n,2)
                  coordinates for point locations
    kwargs      : arguments to pass to Rook, the underlying contiguity class.

    Returns
    -------

    w           : W
                  instance of spatial weights

    Examples
    --------
    >>> import numpy as np
    >>> from libpysal.weights import Voronoi
    >>> np.random.seed(12345)
    >>> points= np.random.random((5,2))*10 + 10
    >>> w = Voronoi(points)
    >>> w.neighbors
    {0: [2, 3, 4], 1: [2], 2: [0, 1, 4], 3: [0, 4], 4: [0, 2, 3]}
    """
    from ..cg.voronoi import voronoi_frames

    region_df, _ = voronoi_frames(points, clip=clip)
    if criterion.lower() == "queen":
        cls = Queen
    elif criterion.lower() == "rook":
        cls = Rook
    else:
        raise ValueError(
            "Contiguity criterion {} not supported. "
            'Only "rook" and "queen" are supported.'.format(criterion)
        )
    return cls.from_dataframe(region_df, **kwargs)


def _from_dataframe(df, **kwargs):
    """
    Construct a voronoi contiguity weight directly from a dataframe. 
    Note that if criterion='rook', this is identical to the delaunay
    graph for the points. 

    If the input dataframe is of any other geometry type than "Point",
    a value error is raised. 

    Parameters
    ----------
    df          :   pandas.DataFrame
                    dataframe containing point geometries for a 
                    voronoi diagram.

    Returns
    -------
    w           :   W
                    instance of spatial weights.
    """
    try:
        x, y = df.geometry.x.values, df.geometry.y.values
    except ValueError:
        raise NotImplementedError(
            "Voronoi weights are only"
            " implemented for point geometries. "
            "You may consider using df.centroid."
        )
    coords = numpy.column_stack((x, y))
    return Voronoi(coords, **kwargs)


Voronoi.from_dataframe = _from_dataframe


def _build(polygons, criterion="rook", ids=None):
    """
    This is a developer-facing function to construct a spatial weights object. 

    Parameters
    ----------
    polygons    : list
                  list of pysal polygons to use to build contiguity
    criterion   : string
                  option of which kind of contiguity to build. Is either "rook" or "queen" 
    ids         : list
                  list of ids to use to index the neighbor dictionary

    Returns
    -------
    tuple containing (neighbors, ids), where neighbors is a dictionary
    describing contiguity relations and ids is the list of ids used to index
    that dictionary. 

    NOTE: this is different from the prior behavior of buildContiguity, which
          returned an actual weights object. Since this just dispatches for the
          classes above, this returns the raw ingredients for a spatial weights
          object, not the object itself. 
    """
    if ids and len(ids) != len(set(ids)):
        raise ValueError(
            "The argument to the ids parameter contains duplicate entries."
        )

    wttype = WT_TYPE[criterion.lower()]
    geo = polygons
    if issubclass(type(geo), FileIO):
        geo.seek(0)  # Make sure we read from the beginning of the file.

    neighbor_data = ContiguityWeightsLists(polygons, wttype=wttype).w

    neighbors = {}
    # weights={}
    if ids:
        for key in neighbor_data:
            ida = ids[key]
            if ida not in neighbors:
                neighbors[ida] = set()
            neighbors[ida].update([ids[x] for x in neighbor_data[key]])
        for key in neighbors:
            neighbors[key] = set(neighbors[key])
    else:
        for key in neighbor_data:
            neighbors[key] = set(neighbor_data[key])
    return (
        dict(
            list(zip(list(neighbors.keys()), list(map(list, list(neighbors.values())))))
        ),
        ids,
    )


def buildContiguity(polygons, criterion="rook", ids=None):
    """
    This is a deprecated function.

    It builds a contiguity W from the polygons provided. As such, it is now
    identical to calling the class constructors for Rook or Queen. 
    """
    # Warn('This function is deprecated. Please use the Rook or Queen classes',
    #        UserWarning)
    if criterion.lower() == "rook":
        return Rook(polygons, ids=ids)
    elif criterion.lower() == "queen":
        return Queen(polygons, ids=ids)
    else:
        raise Exception('Weights criterion "{}" was not found.'.format(criterion))
